# -*- coding: utf-8 -*-
# Custom interpolator script for equal area grid output of the SuperDARN model.
# given data structure is a dictionary = {lat_val: array(time, lon)}
# need an interpolator for each lat_val, then input into an interpolator for
# the outputs vs latitude.

from scipy.interpolate import interp1d
from numpy import NaN, array, unique, zeros, ravel, ndarray
from kamodo_ccmc.readers.reader_utilities import get_slice_idx


def custom_interp(lon_dict, lat_grid, data_dict):
    '''Create a custom interpolator for a given variable for the superdarn
    equal area gridded output.
    Inputs:
        - time_grid: a 1D array of time values in hours.
        - longitude: a dictionary of the form {lat_val: array of lons}
        - lat_grid: a 1D array of latitude values in degrees
        - data_dict: a dictionary of the form {lat_val: array of values}
        Both input dictionaries have arrays with dimensions (time, lon).
    Outputs:
        - An kamodofied interpolator for the given data that accepts
        xvec = [[t1, lon1, lat1], [t2, lon2, lat2], ...].'''

    # initialize lists
    lat_interps, latidx_map = [], []

    def interp_i(lon, lat):
        '''lon, and lat are either 1D arrays or float values.'''
        if isinstance(lat, ndarray):
            out_vals = zeros(lon.shape)
        ulats = unique(lat)
        # retrieve all lon values for given lat_val
        for lat_val in ulats:
            if isinstance(lat, ndarray):
                uidx = [i for i, llat in enumerate(lat) if llat == lat_val]
                position = lon[uidx]
            else:
                position = lon
            # add lat interpolators as needed per lat value
            lat_idxs = get_slice_idx(lat_val, lat_grid)
            for i in lat_idxs:
                if i not in latidx_map:  # add interpolator over longitude
                    latidx_map.append(i)
                    lat_int = interp1d(
                        lon_dict[lat_grid[i]], data_dict[lat_grid[i]],
                        bounds_error=False, fill_value=NaN)  # returns data_val
                    lat_interps.append(lat_int)
            if len(lat_idxs) > 1:
                interp_locations = [latidx_map.index(val) for val in lat_idxs]
                # get values for each latitude values in the index list
                # once transposed, same shape as (len(uidx), len(lat_idx))
                interp_values = array([lat_interps[i](position)
                                       for i in interp_locations]).T
                if isinstance(lat, ndarray):
                    for j, vals in enumerate(interp_values):  # loop lons
                        interp_lat = interp1d(lat_grid[lat_idxs], vals,
                                              bounds_error=False,
                                              fill_value=NaN)
                        out_vals[uidx[j]] = interp_lat(lat_val)
                else:  # only one lon, so interp_values.shape =? lon.shape
                    interp_lat = interp1d(lat_grid[lat_idxs],
                                          ravel(interp_values),
                                          bounds_error=False, fill_value=NaN)
                    out_vals = interp_lat(lat_val)
            else:
                if isinstance(lat, ndarray):
                    interp_location = latidx_map.index(lat_idxs[0])
                    out_vals[uidx] = lat_interps[interp_location](position)
                else:
                    interp_location = latidx_map.index(lat_idxs[0])
                    out_vals = lat_interps[interp_location](position)
        return out_vals

    def total_interp(xvec):
        lon, lat = array(xvec).T
        return interp_i(lon, lat)
    return total_interp



# !/usr/bin/env python
# -----------------------------------------------------------------------------
# $Id: gitm.py,v 1.16 2014/04/09 10:44:28 agburr Exp $
#
# GITM.py, Dan Welling, UMich
#
# Comments: Defines a class for GITM binary output files and modifies the input
#           to improve the data analysis and plotting experience
#
# Contains: class GitmBin    - The class for the GITM binary, which will read
#                              a single GITM output binary
#           def calc_magdi   - Reads a single GITM ion or mag output binary and
#                              computes the magnetic inclination and
#                              declination
#           def calc_magvel  - Reads a single GITM ion output binary and
#                              uses data from the standard GITM output (3DAll)
#                              to compute the ion characteristics in magnetic
#                              coordinates
#           def calc_deg     - Computes and appends latitude and longitude
#                              in degrees from values in radians
#           def calc_lt      - Computes and appends local time in hours from
#                              the universal time for the file and longitude
#           def append_units - Appends unit, descriptive name, and scale
#                              attributes to each known data type
#           def append_data  - Appends a list of data variables to a GitmBin
#                              data structure, where only a limited number of
#                              data variables from that file have been read
#                              in before
#           def calc_tec     - Calculate the VTEC
#           def calc_2dion   - Calculate the 2D ionospheric parameters (VTEC,
#                              hmF2, NmF2)
#
# Updates:
#          Angeline Burrell (AGB) - 1/7/13: Added calc_lt, append_units, and
#                                           calc_magvel
#          AGB - 11/7/13: Improved calc_2dion, added Aaron Ridley's calc_tec
#          AGB - 12/6/13: Added Inclination/Declination calculation
#          Darren De Zeeuw (DDZ) - 06/24/19: Updated code to python3,
#                 Aaron Ridley approved reader for open source use in Kamodo
#          Rebecca Ringuette - 05/14/2021: Removed spacepy dependency and
#                 converted output to a netCDF4 file for increased speed
#                 in later layers. Also removed and adapted calc_magdi,
#                 calc_magvel,
#                 calc_lt for generalized use by other readers. Copied and
#                 adapted calc_tec and calc_2dion for use by other readers.
#                 Removed append_data since converting whole file to a cdf.
# -----------------------------------------------------------------------------


# Global imports:
from glob import glob
import numpy as np
from time import perf_counter
from datetime import datetime, timezone
from netCDF4 import Dataset
from kamodo import Kamodo
from kamodo_ccmc.readers.reader_utilities import Functionalize_Dataset


# replace weird name with standard names in cdf file
gitm_varnames = {'Argon Mixing Ratio': ['r_Ar', 'linear', ''],
                 'Ar Mass Density': ['rho_Ar', 'exponential', 'kg/m**3'],
                 'Methane Mixing Ratio': ['r_CH4', 'linear', ''],
                 'Conduction': ['k', 'linear', 'W/m/K'],
                 'EUV Heating': ['Q_EUV', 'linear', 'K per timestep'],
                 'H Mass Density': ['rho_H', 'exponential', 'kg/m**3'],
                 'H+ Mass Density': ['rho_Hplus', 'exponential', 'kg/m**3'],
                 'H2 Mixing Ratio': ['r_H2', 'linear', ''],
                 'Hydrogen Cyanide Mixing Ratio': ['r_HCN', 'linear', ''],
                 'He Mass Density': ['rho_He', 'exponential', 'kg/m**3'],
                 'He+ Mass Density': ['rho_Heplus', 'exponential', 'kg/m**3'],
                 'Heating Efficiency': ['HeatingEfficiency', 'linear', ''],
                 'Heat Balance Total': ['HeatBalanceTotal', 'linear', ''],
                 'N2 Mass Density': ['rho_N2', 'exponential', 'kg/m**3'],
                 'N2+ Mass Density': ['rho_N2plus', 'exponential', 'kg/m**3'],
                 'N+ Mass Density': ['rho_Nplus', 'exponential', 'kg/m**3'],
                 'N(2D) Mass Density': ['rho_N2D', 'exponential', 'kg/m**3'],
                 'N(2P) Mass Density': ['rho_N2P', 'exponential', 'kg/m**3'],
                 'N(4S) Mass Density': ['rho_N4S', 'exponential', 'kg/m**3'],
                 'N2 Mixing Ratio': ['r_N2', 'linear', ''],
                 'NO Mass Density': ['rho_NO', 'exponential', 'kg/m**3'],
                 'NO+ Mass Density': ['rho_NOplus', 'exponential', 'kg/m**3'],
                 'O2 Mass Density': ['rho_O2', 'exponential', 'kg/m**3'],
                 'O(1D) Mass Density': ['rho_O1D', 'exponential', 'kg/m**3'],
                 'O2+ Mass Density': ['rho_O2plus', 'exponential', 'kg/m**3'],
                 'O(2D) Mass Density': ['rho_O2D', 'exponential', '1/m**3'],
                 'O+(2P) Mass Density': ['rho_Oplus2P', 'exponential',
                                         'kg/m**3'],
                 'O(3P) Mass Density': ['rho_O3P', 'exponential', 'kg/m**3'],
                 'O+(4SP) Mass Density': ['rho_Oplus4SP', 'exponential',
                                          'kg/m**3'],
                 'Radiative Cooling': ['L_Rad', 'linear', ''],
                 'Neutral Density': ['rho', 'exponential', 'kg/m**3'],
                 'T_n': ['T_n', 'linear', 'K'],
                 'vi_east': ['vi_east', 'linear', 'm/s'],
                 'vi_north': ['vi_north', 'linear', 'm/s'],
                 'vi_up': ['vi_up', 'linear', 'm/s'],
                 'vn_east': ['vn_east', 'linear', 'm/s'],
                 'vn_north': ['vn_north', 'linear', 'm/s'],
                 'vn_up': ['vn_up', 'linear', 'm/s'],
                 'v_N2_up': ['v_N2_up', 'linear', 'm/s'],
                 'v_N(4S)_up': ['v_N4S_up', 'linear', 'm/s'],
                 'v_N_up': ['v_N_up', 'linear', 'm/s'],
                 'v_O2_up': ['v_O2_up', 'linear', 'm/s'],
                 'v_O(3P)_up': ['v_O3P_up', 'linear', 'm/s'],
                 'v_He_up': ['v_He_up', 'linear', 'm/s'],
                 '[e-]': ['rho_e', 'linear', '1/m**3'],
                 'Electron Average Energy': ['ElectronAverageEnergy', 'linear',
                                             'J'],
                 'T_e': ['T_e', 'linear', 'K'],
                 'T_i': ['T_i', 'linear', 'K'],
                 'Solar Zenith Angle': ['SolarZenithAngle', 'linear',
                                        'radians'],
                 'CO2 Mass Density': ['rho_CO2', 'exponential', 'kg/m**3'],
                 'DivJu FL': ['DivJuFL', '', ''],
                 'DivJuAlt': ['DivJuAlt', 'linear', ''],
                 'Electron Energy Flux': ['ElectronEnergyFlux', 'exponential',
                                          'J/m**2'],
                 'Field Line Length': ['Field Line Length', 'linear', 'm'],
                 'sigma_P': ['sigma_P', 'linear', 'S/m'],
                 'Sigma_P': ['Sigma_P', 'linear', 'S/m'],
                 'sigma_H': ['sigma_H', 'linear', 'S/m'],
                 'Potential': ['V', 'linear', 'V'],
                 'Sigma_H': ['Sigma_H', 'linear', 'S/m'],
                 'Region 2 Current': ['I_R2', 'linear', 'A/m**2'],
                 'Region 1 Current': ['I_R1', 'linear', 'A/m**2'],
                 'Ed1': ['Ed1', 'linear', ''],
                 'Ed2': ['Ed2', 'linear', ''],
                 'Solar Local Time': ['SolarLocalTime', 'linear', 'h'],
                 'Vertical Electric Field': ['E_up', 'linear', 'V/m'],
                 'Eastward Electric Field': ['E_east', 'linear', 'V/m'],
                 'Northward Electric Field': ['E_north', 'linear', 'V/m'],
                 'Electric Field Magnitude': ['E_mag', 'linear', 'V/m'],
                 'Vertical Magnetic Field': ['B_up', 'linear', 'nT'],
                 'Eastward Magnetic Field': ['B_east', 'linear', 'nT'],
                 'Northward Magnetic Field': ['B_north', 'linear', 'nT'],
                 'Magnetic Field Magnitude': ['B_mag', 'linear', 'nT'],
                 'Magnetic Latitude': ['MagLat', 'linear', 'deg'],
                 'Magnetic Longitude': ['MagLon', 'linear', 'deg'],
                 'g': ['g', 'linear', 'm/s**2'],
                 'GradP_east (P_i + P_e)': ['GradP_east', 'linear', 'Pa/m'],
                 'GradP_north (P_i + P_e)': ['GradP_north', 'linear', 'Pa/m'],
                 'GradP_up (P_i + P_e)': ['GradP_up', 'linear', 'Pa/m'],
                 'nu_in': ['nu_in', 'linear', '1/s'],
                 'Chemical Heating Rate': ['ChemicalHeatingRate', 'linear',
                                           ''],
                 'Total Absolute EUV': ['TotalAbsoluteEUV', 'linear',
                                        'K per timestep'],
                 'O Cooling': ['L_O', 'linear', 'K per timestep'],
                 'Joule Heating': ['Q_Joule', 'linear', 'K per timestep'],
                 'Auroral Heating': ['Q_Auroral', 'linear', 'K per timestep'],
                 'Photoelectron Heating': ['Q_PhotoE', 'linear',
                                           'K per timestep'],
                 'Eddy Conduction': ['k_eddy', 'linear', ''],
                 'Adiabatic Eddy Conduction': ['k_adiabaticeddy', 'linear',
                                               ''],
                 'NO Cooling': ['L_NO', 'linear', 'K per timestep'],
                 'Molecular Conduction': ['k_molecular', 'linear', ''],
                 'max_electron_density': ['NmF2', 'linear', ''],
                 'max_electron_density_height': ['hmF2', 'linear', 'km'],
                 'Vertical TEC': ['TEC', 'linear', '10**16/m**2'],
                 'HeatFlux_Joule': ['phi_qJoule', 'linear', 'W/m**2'],
                 'HeatFlux': ['phi_q', 'linear', 'W/m**2'],
                 'HeatFlux_EUV': ['phi_qEUV', 'linear', 'W/m**2'],
                 'NO CoolingFlux': ['phi_qNOCooling', 'linear', 'W/m**2']}

# if not always one value, deal with this in kamodo wrapper
name_dict = {"Altitude": "Altitude", "Ar Mixing Ratio": "Argon Mixing Ratio",
             "Ar": "Ar Mass Density",
             "CH4 Mixing Ratio": "Methane Mixing Ratio",
             "Conduction": "Conduction", "EuvHeating": "EUV Heating",
             "H": "H Mass Density", "H!U+!N": "H+ Mass Density",
             "H2 Mixing Ratio": "H2 Mixing Ratio",
             "HCN Mixing Ratio": "Hydrogen Cyanide Mixing Ratio",
             "He": "He Mass Density", "He!U+!N": "He+ Mass Density",
             "Heating Efficiency": "Heating Efficiency",
             "Heat Balance Total": "Heat Balance Total",
             "Latitude": "Latitude", "Longitude": "Longitude",
             "N!D2!N": "N2 Mass Density", "N!D2!U+!N": "N2+ Mass Density",
             "N!U+!N": "N+ Mass Density", "N(!U2!ND)": "N(2D) Mass Density",
             "N(!U2!NP)": "N(2P) Mass Density",
             "N(!U4!NS)": "N(4S) Mass Density",
             "N2 Mixing Ratio": "N2 Mixing Ratio", "NO": "NO Mass Density",
             "NO!U+!N": "NO+ Mass Density", "O!D2!N": "O2 Mass Density",
             "O(!U1!ND)": "O(1D) Mass Density",
             "O!D2!U+!N": "O2+ Mass Density",
             "O(!U2!ND)!": "O(2D) Mass Density",
             "O(!U2!ND)!U+!N": "O(2D) Mass Density",
             "O(!U2!NP)!U+!N": "O+(2P) Mass Density",
             "O(!U3!NP)": "O(3P) Mass Density",
             "O_4SP_!U+!N": "O+(4SP) Mass Density",
             "RadCooling": "Radiative Cooling",
             "Rho": "Neutral Density", "Temperature": "T_n",
             "V!Di!N (east)": "vi_east",
             "V!Di!N (north)": "vi_north", "Vi (north) (m/s)": "vi_north",
             "V!Di!N (up)": "vi_up", "Vi (up) (m/s)": "vi_up",
             "V!Dn!N (east)": "vn_east",
             "Vn (east) (m/s)": "vn_east", "V!Dn!N (north)": "vn_north",
             "Vn (north) (m/s)": "vn_north", "V!Dn!N (up)": "vn_up",
             "Vn (up) (m/s)": "vn_up",
             "V!Dn!N (up,N!D2!N              )": "v_N2_up",
             "V!Dn!N (up,N(!U4!NS)           )": "v_N(4S)_up",
             "V!Dn!N (up,NO                  )": "v_N_up",
             "V!Dn!N (up,O!D2!N              )": "v_O2_up",
             "V!Dn!N (up,O(!U3!NP)           )": "v_O(3P)_up",
             "V!Dn!N (up,He                  )": "v_He_up",
             "e-": "[e-]",
             "Electron_Average_Energy": "Electron Average Energy",
             "eTemperature": "T_e", "iTemperature": "T_i",
             "Solar Zenith Angle": "Solar Zenith Angle",
             "Vertical TEC": "Vertical TEC", "CO!D2!N": "CO2 Mass Density",
             "DivJu FL": "DivJu FL", "DivJuAlt": "DivJuAlt",
             "Electron_Energy_Flux": "Electron Energy Flux",
             "FL Length": "Field Line Length",
             "Pedersen FL Conductance": "sigma_P",
             "Pedersen Conductance": "Sigma_P",
             "Hall FL Conductance": "sigma_H",
             "Potential": "Potential", "Hall Conductance": "Sigma_H",
             "Je2": "Region 2 Current", "Je1": "Region 1 Current",
             "Ed1": "Ed1", "Ed2": "Ed2", "LT": "Solar Local Time",
             "Local Time": "Solar Local Time",
             "E.F. Vertical": "Vertical Electric Field",
             "E.F. East": "Eastward Electric Field",
             "E.F. North": "Northward Electric Field",
             "E.F. Magnitude": "Electric Field Magnitude",
             "B.F. Vertical": "Vertical Magnetic Field",
             "B.F. East": "Eastward Magnetic Field",
             "B.F. North": "Northward Magnetic Field",
             "B.F. Magnitude": "Magnetic Field Magnitude",
             "Magnetic Latitude": "Magnetic Latitude",
             "Magnetic Longitude": "Magnetic Longitude", "dLat": "Latitude",
             "dLon": "Longitude",
             "Gravity": "g", "PressGrad (east)": "GradP_east (P_i + P_e)",
             "PressGrad (north)": "GradP_north (P_i + P_e)",
             "PressGrad (up)": "GradP_up (P_i + P_e)",
             "IN Collision Freq": "nu_in",
             "Chemical Heating": "Chemical Heating Rate",
             "Total Abs EUV": "Total Absolute EUV",
             "O Cooling": "O Cooling", "Joule Heating": "Joule Heating",
             "Auroral Heating": "Auroral Heating",
             "Photoelectron Heating": "Photoelectron Heating",
             "Eddy Conduction": "Eddy Conduction",
             "Eddy Adiabatic Conduction": "Adiabatic Eddy Conduction",
             "NO Cooling": "NO Cooling",
             "Molecular Conduction": "Molecular Conduction",
             'NmF2': 'max_electron_density',
             'hmF2': "max_electron_density_height",
             'VTEC': "Vertical TEC",
             'AltIntJouleHeating (W/m2)': 'HeatFlux_Joule',
             'AltIntHeatingTransfer (W/m2)': 'HeatFlux',
             'AltIntEuvHeating (W/m2)': 'HeatFlux_EUV',
             'AltIntNOCooling (W/m2)': 'NO CoolingFlux'}

unit_dict = {"Altitude": "m", "Ar Mixing Ratio": "", "Ar": "kg/m**3",
             "CH4 Mixing Ratio": "", "Conduction": "W/m/K",
             "EuvHeating": "K per timestep",
             "H": "kg/m**3", "H!U+!N": "kg/m**3", "H2 Mixing Ratio": "",
             "HCN Mixing Ratio": "", "He": "kg/m**3", "He!U+!N": "kg/m**3",
             "Heating Efficiency": "", "Heat Balance Total": "",
             "Latitude": "radians",
             "Longitude": "radians", "N!D2!N": "kg/m**3",
             "N!D2!U+!N": "kg/m**3",
             "N!D2!U+!N            (/m3)": "1/m**3", "N!U+!N": "kg/m**3",
             "N(!U2!ND)": "kg/m**3", "N(!U2!NP)": "kg/m**3",
             "N(!U4!NS)": "kg/m**3",
             "N2 Mixing Ratio": "", "NO": "kg/m**3", "NO!U+!N": "kg/m**3",
             "O!D2!N": "kg/m**3", "O(!U1!ND)": "kg/m**3",
             "O!D2!U+!N": "kg/m**3",
             "O(!U2!ND)!": "1/m**3", "O(!U2!ND)!U+!N": "1/m**3",
             "O(!U2!NP)!U+!N": "kg/m**3",
             "O(!U2!NP)!U+!N": "kg/m**3", "O(!U3!NP)": "kg/m**3",
             "O_4SP_!U+!N": "kg/m**3",
             "RadCooling": "", "Rho": "kg/m**3", "Temperature": "K",
             "V!Di!N (east)": "m/s",
             "Vi (east) (m/s)": "m/s", "V!Di!N (north)": "m/s",
             "Vi (north) (m/s)": "m/s",
             "V!Di!N (up)": "m/s", "Vi (up) (m/s)": "m/s",
             "V!Dn!N (east)": "m/s",
             "Vn (east) (m/s)": "m/s", "V!Dn!N (north)": "m/s",
             "Vn (north) (m/s)": "m/s",
             "V!Dn!N (up)": "m/s", "Vn (up) (m/s)": "m/s",
             "V!Dn!N (up,N!D2!N              )": "m/s",
             "V!Dn!N (up,N(!U4!NS)           )": "m/s",
             "V!Dn!N (up,NO                  )": "m/s",
             "V!Dn!N (up,O!D2!N              )": "m/s",
             "V!Dn!N (up,O(!U3!NP)           )": "m/s",
             "V!Dn!N (up,He                  )": "m/s",
             "e-": "1/m**3", "Electron_Average_Energy": "J",
             "eTemperature": "K",
             "iTemperature": "K", "LT": "h", "Local Time": "h",
             "Solar Zenith Angle": "radians",
             "Vertical TEC": "10**16/m**2", "CO!D2!N": "kg/m**3",
             "DivJu FL": "", "DivJuAlt": "",
             "Electron_Energy_Flux": "J/m**2", "FL Length": "m",
             "Pedersen FL Conductance": "S/m",
             "dLon": "deg", "Pedersen Conductance": "S/m", "dLat": "deg",
             "Hall FL Conductance": "S/m", "Potential": "V",
             "Hall Conductance": "S/m",
             "Je2": "A/m**2", "Je1": "A/m**2", "Magnetic Longitude": "deg",
             "E.F. Vertical": "V/m", "E.F. East": "V/m", "E.F. North": "V/m",
             "E.F. Magnitude": "V/m",
             "B.F. Vertical": "nT", "B.F. East": "nT", "B.F. North": "nT",
             "B.F. Magnitude": "nT",
             "Magnetic Latitude": "deg", "Ed1": "", "Ed2": "",
             "Gravity": "m/s**2",
             "PressGrad (east)": "Pa/m", "Joule Heating": "K per timestep",
             "PressGrad (north)": "Pa/m", "O Cooling": "K per timestep",
             "PressGrad (up)": "Pa/m", "Total Abs EUV": "K per timestep",
             "IN Collision Freq": "1/s", "Chemical Heating": "",
             "Auroral Heating": "K per timestep",
             "Photoelectron Heating": "K per timestep",
             "Eddy Conduction": "", "Eddy Adiabatic Conduction": "",
             "NO Cooling": "K per timestep", "Molecular Conduction": "",
             'NmF2': "", 'hmF2': "km", 'VTEC': "10**16/m**2",
             'AltIntJouleHeating (W/m2)': 'W/m**2',
             'AltIntHeatingTransfer (W/m2)': 'W/m**2',
             'AltIntEuvHeating (W/m2)': 'W/m**2',
             'AltIntNOCooling (W/m2)': 'W/m**2'}

scale_dict = {"Altitude": "linear", "Ar Mixing Ratio": "linear",
              "Ar": "exponential", "CH4 Mixing Ratio": "linear",
              "Conduction": "linear", "EuvHeating": "linear",
              "H": "exponential", "H!U+!N": "exponential",
              "H2 Mixing Ratio": "linear", "HCN Mixing Ratio": "linear",
              "He": "exponential", "He!U+!N": "exponential",
              "Heating Efficiency": "linear", "DivJuAlt": "linear",
              "Heat Balance Total": "linear", "Latitude": "linear",
              "Longitude": "linear", "N!D2!N": "exponential",
              "N!D2!U+!N": "exponential",
              "N!U+!N": "exponential", "N(!U2!ND)": "exponential",
              "N(!U2!NP)": "exponential",
              "N(!U4!NS)": "exponential", "N2 Mixing Ratio": "linear",
              "NO": "exponential",
              "NO!U+!N": "exponential", "O!D2!N": "exponential",
              "O!D2!U+!N": "exponential",
              "O(!U2!ND)!": "exponential", "O(!U1!ND)": "exponential",
              "O(!U2!ND)!U+!N": "exponential", "CO!D2!N": "exponential",
              "O(!U2!NP)!U+!N": "exponential", "DivJu FL": "",
              "O(!U2!NP)!U+!N": "exponential", "O(!U3!NP)": "exponential",
              "O_4SP_!U+!N": "exponential", "RadCooling": "linear",
              "Rho": "exponential", "Temperature": "linear",
              "V!Di!N (east)": "linear",
              "Vi (east)": "linear", "V!Di!N (north)": "linear",
              "Vi (north) (m/s)": "linear",
              "V!Di!N (up)": "linear", "Vi (up) (m/s)": "linear",
              "V!Dn!N (east)": "linear",
              "Vn (east) (m/s)": "linear", "V!Dn!N (north)": "linear",
              "Vn (north) (m/s)": "linear",
              "V!Dn!N (up)": "linear", "Vn (up) (m/s)": "linear",
              "V!Dn!N (up,N!D2!N              )": "linear",
              "V!Dn!N (up,N(!U4!NS)           )": "linear",
              "V!Dn!N (up,NO                  )": "linear",
              "V!Dn!N (up,O!D2!N              )": "linear",
              "V!Dn!N (up,O(!U3!NP)           )": "linear",
              "V!Dn!N (up,He                  )": "linear",
              "e-": "linear",
              # "e-                   (/m3)": "linear",
              "Electron_Average_Energy": "linear",
              "eTemperature": "linear", "iTemperature": "linear",
              "Solar Zenith Angle": "linear", "Vertical TEC": "linear",
              "Electron_Energy_Flux": "exponential",
              "FL Length": "linear", "Pedersen FL Conductance": "linear",
              "Hall Conductance": "linear", "Potential": "linear",
              "Hall FL Conductance": "linear", "dLon": "linear",
              "Pedersen Conductance": "linear", "Je2": "linear",
              "Je1": "linear", "Ed1": "linear", "Ed2": "linear",
              "E.F. Vertical": "linear", "E.F. East": "linear",
              "E.F. North": "linear", "E.F. Magnitude": "linear",
              "B.F. Vertical": "linear", "B.F. East": "linear",
              "B.F. North": "linear", "B.F. Magnitude": "linear",
              "Magnetic Latitude": "linear", "LT": "linear",
              "Local Time": "linear", "Magnetic Longitude": "linear",
              "dLat": "linear",
              "Gravity": "linear", "PressGrad (east)": "linear",
              "PressGrad (north)": "linear", "PressGrad (up)": "linear",
              "IN Collision Freq": "linear", "Chemical Heating": "linear",
              "Total Abs EUV": "linear", "O Cooling": "linear",
              "Joule Heating": "linear", "Auroral Heating": "linear",
              "Photoelectron Heating": "linear", "NO Cooling": "linear",
              "Eddy Conduction": "linear", "Molecular Conduction": "linear",
              "Eddy Adiabatic Conduction": "linear",
              'NmF2': "linear", 'hmF2': "linear", 'VTEC': 'linear',
              'AltIntJouleHeating (W/m2)': 'linear',
              'AltIntHeatingTransfer (W/m2)': 'linear',
              'AltIntEuvHeating (W/m2)': 'linear',
              'AltIntNOCooling (W/m2)': 'linear'}


def _read(varlist, attrs, newfile=True):
    '''
    Read binary file.
    '''

    from re import sub
    from struct import unpack
    import sys

    # Read data and header info
    coords, variables = {}, {}
    f = open(attrs['file'], 'rb')

    # Using the first FORTRAN header, determine endian.
    # Default is little.
    attrs['endian'] = 'little'
    endChar = '>'
    rawRecLen = f.read(4)
    if len(rawRecLen) < 4:
        print("GitmBin ERROR: empty file [", attrs['file'], "]")
        sys.exit(1)
    recLen = (unpack(endChar+'l', rawRecLen))[0]
    if (recLen > 10000) or (recLen < 0):
        # Ridiculous record length implies wrong endian.
        attrs['endian'] = 'big'
        endChar = '<'
        recLen = (unpack(endChar+'l', rawRecLen))[0]

    # Read version; read fortran footer+header.
    attrs['version'] = unpack(endChar+'d', f.read(recLen))[0]
    (oldLen, recLen) = unpack(endChar+'2l', f.read(8))

    # Read grid size information.
    (attrs['nLon'], attrs['nLat'], attrs['nAlt']) = \
        unpack(endChar+'lll', f.read(recLen))
    (oldLen, recLen) = unpack(endChar+'2l', f.read(8))

    # Read number of variables.
    attrs['nVars'] = unpack(endChar+'l', f.read(recLen))[0]
    (oldLen, recLen) = unpack(endChar+'2l', f.read(8))

    # Collect variable names.
    var = []
    for i in range(attrs['nVars']):
        var.append(unpack(endChar+'%is' % (recLen), f.read(recLen))[0])
        (oldLen, recLen) = unpack(endChar+'2l', f.read(8))

    # Extract time.
    (yy, mm, dd, hh, mn, ss, ms) = unpack(endChar+'lllllll', f.read(recLen))
    # self._time=dt.datetime(yy,mm,dd,hh,mn,ss,ms/1000)
    attrs['time'] = datetime(yy, mm, dd, hh, mn, ss, ms).replace(
        tzinfo=timezone.utc).isoformat(sep=' ')
    (oldLen) = unpack(endChar+'l', f.read(4))

    # Read the rest of the data. Have to read through whole file.
    nTotal = attrs['nLon']*attrs['nLat']*attrs['nAlt']
    for val in var:  # loops through all variable data stored in file
        # Trim variable names.
        v = sub('\[|\]', '', val.decode('utf-8')).strip()
        s = unpack(endChar+'l', f.read(4))[0]
        # Test to see if this variable is desired
        gvar = True
        if len(varlist) > 0:
            try:
                varlist.index(v)
            except ValueError:  # include coordinate data or data from new file
                if ((v != 'Altitude' and v != 'Longitude' and
                     v != 'Latitude') or not newfile):
                    gvar = False
        # Unpack the data and save, if desired
        temp = unpack(endChar+'%id' % (nTotal), f.read(s))
        if gvar:
            # if a coordinate, only take unique values to speed up calculations
            if v == 'Longitude':
                coords['lon'] = np.unique(np.array(temp))*180.0/np.pi
            elif v == 'Latitude':
                coords['lat'] = np.unique(np.array(temp))*180.0/np.pi
            elif v == 'Altitude':
                # km, need for height integration
                coords['height'] = np.unique(np.array(temp))/1000.
            else:  # put in height, lat, lon order, noting fortran ordering
                variables[v] = np.transpose(np.array(temp).reshape(
                    (attrs['nLon'], attrs['nLat'], attrs['nAlt']),
                    order='F'), [2, 1, 0])
        f.read(4)
    return coords, variables, attrs


def append_units(variables):
    '''
    Append units, descriptive names, and plot scaling (e.g. linear,
    exponential) to the attributes of known data types
    '''

    var_dict = {}  # to store names, units, etc
    for k in list(variables.keys()):
        if type(variables[k]) is np.ndarray:
            nk = k
            # Temporary fix for misspelled key (9/30/13)
            if nk.find("Heaing Efficiency") >= 0:
                nk = "Heating Efficiency"
            elif nk.find("EUV Heating") >= 0:
                nk = "EuvHeating"
            elif nk.find("Rho (kg/m3)") >= 0:
                nk = "Rho"
            elif nk.find("Neutral Temperature (K)") >= 0:
                nk = 'Temperature'
            elif nk.find("Vn (up) (m/s)") >= 0:
                nk = 'V!Dn!N (up)'
            elif nk.find("Vi (east) (m/s)") >= 0:
                nk = "V!Di!N (east)"
            # DDZ added 2020-08-31
            if nk == "Magnetic latitude":
                nk = "Magnetic Latitude"

            try:
                # print ('registering',k)
                var_dict = register_name(k, nk, unit_dict, scale_dict,
                                         name_dict, var_dict)
            except:
                if k.split()[0] in name_dict:
                    nk = k.split()[0]
                    if "(/m3)" in k:
                        # print 'found (/m3) in key'
                        unit_dict[k] = "m^{-3}"
                        name_dict[k] = name_dict[nk]
                        scale_dict[k] = scale_dict[nk]
                        var_dict = register_name(k, k, unit_dict, scale_dict,
                                                 name_dict, var_dict)
                    else:
                        raise
                else:
                    raise
    return var_dict


def register_name(k, nk, unit_dict, scale_dict, name_dict, var_dict):
    # Different versions of GITM differ in header capitalization
    if nk not in name_dict:
        # Try to capitalize or lowercase the key
        if nk == nk.capitalize():
            nk = k.lower()
        else:
            nk = k.capitalize()
    var_dict[k] = [name_dict[nk], scale_dict[nk], unit_dict[nk]]
    return var_dict


def calc_tec(coords, attrs, variables, verbose=False):
    '''
    A routine to calculate the 2D VTEC.
    To perform these calculations, electron density ("e-") must be one of
    the available data types.
    '''
    import scipy.integrate as integ

    if 'e-' in variables.keys() or 'e-                   (/m3)' in\
            variables.keys():
        if 'e-' in variables.keys():
            e_key = 'e-'
        elif 'e-                   (/m3)' in variables.keys():
            e_key = 'e-                   (/m3)'
        temp = np.array(variables[e_key] * 1.0e-16)
        variables['VTEC'] = np.zeros((attrs['nLat'], attrs['nLon']),
                                     dtype=np.float64)
        for ilon in range(attrs['nLon']):
            for ilat in range(attrs['nLat']):
                # Integrate electron density over altitude, not including
                # ghost cells
                vtec = integ.simps(temp[2:-2, ilat, ilon],
                                   coords['height'][2:-2]*1000, "avg")
                variables['VTEC'][ilat, ilon] = vtec
    elif verbose:
        print('e-!', variables.keys())
    return variables


def calc_2dion(coords, attrs, variables, verbose=False):
    '''
    A routine to calculate the 2D ionospheric parameters: hmF2, NmF2.
    To perform these calculations, electron density ("e-") must be one of
    the available data types.
    '''
    from scipy.interpolate import interp1d
    from scipy.signal import argrelextrema

    if 'e-' in variables.keys() or 'e-                   (/m3)' in\
            variables.keys():
        if 'e-' in variables.keys():
            e_key = 'e-'
        elif 'e-                   (/m3)' in variables.keys():
            e_key = 'e-                   (/m3)'
        variables['NmF2'] = np.zeros((attrs['nLat'], attrs['nLon']),
                                     dtype=np.float64)
        variables['hmF2'] = np.zeros((attrs['nLat'], attrs['nLon']),
                                     dtype=np.float64)
        alt = np.linspace(min(coords['height'][2:-2]),
                          max(coords['height'][2:-2]), 1000)
        ilon, ilat = coords['lon'], coords['lat']
        for ilon in range(attrs['nLon']):
            for ilat in range(attrs['nLat']):

                # Interpolate over the electron density altitude profile
                eprof = interp1d(coords['height'][2:-2],
                                 variables[e_key][2:-2, ilat, ilon],
                                 kind="cubic")
                edens = eprof(alt)

                # Find the local maxima of the electron density profile
                emax = argrelextrema(edens, np.greater)
                emax_list = list(emax[0])
                saddle = False
                if len(emax_list) == 0:
                    # No local maxima were found, try identifying
                    # saddle points
                    saddle = True
                elif len(emax_list) == 1:   # emax_list is a number
                    if max(edens) == edens[emax_list[0]]:
                        # Only one maxima exists and is realistic. Sometimes
                        # a profile will have inflection points instead of
                        # local maxima and this can confuse the routine
                        variables['NmF2'][ilat, ilon] = edens[emax_list[0]]
                        variables['hmF2'][ilat, ilon] = alt[emax_list[0]]
                    else:
                        saddle = True
                elif alt[emax_list[-1]] < 120.0:
                    saddle = True
                else:
                    # More than one maxima exists.  Seperate hmF2 from hmF1
                    # and spurious local maxima
                    NmF2 = list(edens[emax_list])
                    HmF2 = list(alt[emax_list])

                    # If the global maximum is over 200 km,
                    # this is the F2 peak
                    eindex = NmF2.index(max(NmF2))
                    if HmF2[eindex] <= 200.0 and HmF2[eindex] == min(HmF2):
                        # The global max may be the F1 peak, see if the
                        # secondary (or lessor) maxima is at the upper
                        # limit of the model.  If so, remove this point
                        # from consideration
                        if max(HmF2) > coords['height'][-5]:
                            eindex = HmF2.index(max(HmF2))
                            emax_list.pop(eindex)
                            NmF2.pop(eindex)
                            HmF2.pop(eindex)
                            eindex = NmF2.index(max(NmF2))

                        if len(emax_list) > 1:
                            # If there are multiple maxima after the upper
                            # boundary has been removed, choose the largest
                            # maxima above 200 km since the hmF1 is often
                            # larger than the hmF2
                            emax_list.pop(eindex)
                            NmF2.pop(eindex)
                            eindex = NmF2.index(max(NmF2))

                    # Set the hmF2 and NmF2 (#emax_list is a number)
                    variables['NmF2'][ilat, ilon] = edens[emax_list[eindex]]
                    variables['hmF2'][ilat, ilon] = alt[emax_list[eindex]]

                if saddle:
                    # It is difficult to find saddle points.  Examine
                    # the rate of change of density
                    delta_alt = alt[1] - alt[0]  # Equally spaced
                    edens_dot = np.diff(edens) / delta_alt

                    # Identify inflection points by looking for
                    # minima in the derivative of electron density.
                    edot_min = argrelextrema(edens_dot, np.less)
                    emin_list = list(edot_min[0])
                    edens_min = list(edens[emin_list])
                    emax = np.max(edens_min)
                    eindex = emin_list[edens_min.index(emax)]

                    # Assign the inflection with the largest
                    # electron density to the ion peak
                    # emax_list is a number
                    variables['NmF2'][ilat, ilon] = edens[eindex]
                    variables['hmF2'][ilat, ilon] = alt[eindex]
    elif verbose:
        print('e- not found', variables.keys())
    return variables


def GitmBin(filename, flag_2D, verbose=False):
    '''
    Object to open, manipulate and visualize 1-3 dimensional GITM output
    stored in binary format.  Object inherits from spacepy.pybats.PbData; see
    that documentation for general information on how these objects work.

    GITM index ordering is [lon, lat, altitude]; data arrays read from file
    will always be of the same shape and size.

    kwargs may be specified for:
    magfile = 3DION or 3DMAG file, allows computation of velocities in magnetic
              coordinates
    varlist = list of variable keys.  Will limit the variables read in to those
              listed.  Time and position will always be read in.  If the list
              is empty, all variables will be read in.
    '''
    varlist = []
    attrs = {'file': filename}

    # Load the GITM data
    coords, variables, attrs = _read(varlist, attrs)
    if not flag_2D:
        variables = calc_tec(coords, attrs, variables, verbose=verbose)
        variables = calc_2dion(coords, attrs, variables, verbose=verbose)
    var_dict = append_units(variables)
    return filename, coords, variables, var_dict, attrs


def gitm_toCDF(bin_file, coords, variables, var_dict, attrs):
    '''Prepare and write data into netcdf files.'''

    # perform longitude shift, careful of extra values on ends
    # this logic is different because the lon grid doesn't start at zero
    #   or end at 360. In some cases it goes farther on both ends.
    lon_le180 = list(np.where((coords['lon'] <= 180.) &
                              (coords['lon'] >= 0.))[0])
    lon_ge180 = list(np.where((coords['lon'] >= 180.) &
                              (coords['lon'] < 360.))[0])
    if 180. not in coords['lon']:
        lon_ge180.insert(0, lon_le180[-1])
        lon_le180.append(lon_ge180[1])
    full_lon = np.append(coords['lon'][lon_ge180] - 360.,
                         coords['lon'][lon_le180])  # keep separate
    new_lon = np.append(coords['lon'][lon_ge180] - 360.,
                        coords['lon'][lon_le180])  # in memory
    new_lon[0], new_lon[-1] = -180., 180.  # begins at -180, ends at +180.
    coords['lon'] = new_lon  # use corrected grid for file

    # prepare for interpolation at +/- 180 deg lon, possible also latitude
    # interpolation is done after lon wrapping, so wrapped lon grid here
    ko = Kamodo()
    ko.variables = {'TMP': {}}
    ko._registered = 0  # need full lon and lat grids
    coord_dict = {'lon': {'units': 'deg', 'data': full_lon},
                  'lat': {'units': 'deg', 'data': coords['lat']}}

    # check for and prepare for latitude wrapping or interpolation
    lat_check = False
    if coords['lat'].max() < 90.:
        coords['lat'] = np.insert(coords['lat'], 0, -90.)
        coords['lat'] = np.append(coords['lat'], 90.)
        coord_dict['lat']['data'] = coords['lat']
        print('Performing scalar averaging at the poles.')
    elif coords['lat'].max() > 90.:
        old_lat = coords['lat']
        lat_idx = list(np.where((old_lat > -90.) & (old_lat < 90.))[0])
        lat_idx = [min(lat_idx) - 1] + lat_idx + [max(lat_idx) + 1]  # add ends
        coord_dict['lat']['data'] = old_lat[lat_idx]
        coords['lat'] = old_lat[lat_idx]  # corrected lat grid for files
        coords['lat'][0], coords['lat'][-1] = -90., 90.
        lat_check = True  # signal to perform interpolation at poles
        print('Interpolating data at the poles.')

    # start new output object
    cdf_filename = bin_file.replace('.bin', '.nc')
    data_out = Dataset(cdf_filename, 'w', format='NETCDF4')
    # csv list of files
    data_out.file = bin_file
    data_out.model = 'GITM'
    data_out.version = attrs['version']
    data_out.endian = attrs['endian']
    data_out.filedate = attrs['filedate']
    for dim in coords.keys():  # register dimensions
        if coords[dim].size == 1 and dim != 'time':
            continue  # skip height for 2D variables
        # create dimension
        new_dim = data_out.createDimension(dim, coords[dim].size)
        # create variable
        new_var = data_out.createVariable(dim, np.float32, dim)
        new_var[:] = coords[dim]  # store data for dimension in variable
        if dim == 'height':
            units = 'km'
        if dim == 'time':
            units = 'hr'
        else:
            units = 'deg'
        new_var.datascale, new_var.units = 'linear', units

    # copy over variables to file
    # var_3D = ['TEC', 'NmF2', 'hmF2', 'SolarLocalTime', 'SolarZenithAngle',
    #          'phi_qJoule', 'phi_q', 'phi_qEUV', 'phi_qNOCooling']
    for variable_name in variables.keys():
        new_name = gitm_varnames[var_dict[variable_name][0]][0]
        # remove height dimension for 2D data
        if 1 in variables[variable_name].shape:
            variables[variable_name] = np.squeeze(variables[variable_name])
        if len(variables[variable_name].shape) == 3:
            new_var = data_out.createVariable(new_name, np.float32,
                                              ('lon', 'lat', 'height'))
            if 'height' not in coord_dict.keys():
                coord_dict['height'] = {'units': 'km',
                                        'data': coords['height']}
        elif len(variables[variable_name].shape) == 2:
            new_var = data_out.createVariable(new_name, np.float32,
                                              ('lon', 'lat'))
            if 'height' in coord_dict.keys():
                del coord_dict['height']
        # 3D: (h,lat,lon) -> (lon,lat,h), 2D: (lat,lon) -> (lon,lat)
        tmp = variables[variable_name].T

        # check for and perform latitude wrapping or prepare for interpolation
        if lat_check:
            tmp = tmp[:, lat_idx]  # chop off extra latitude values
        elif tmp.shape[1] < len(coords['lat']):
            spole = np.mean(tmp[:, 0], axis=0)  # avg over lons near pole
            npole = np.mean(tmp[:, -1], axis=0)  # other pole
            tmp = np.insert(tmp, tmp.shape[1]-1, npole, axis=1)  # append
            tmp = np.insert(tmp, 0, spole, axis=1)  # insert on lat axis
        new_data = tmp[lon_ge180 + lon_le180]  # wrap in longitude

        # replace lon borders with proper interpolated boundary values
        ko = Functionalize_Dataset(ko, coord_dict, 'TMP',
                                   {'data': new_data, 'units': ''}, True,
                                   'GDZsph')
        new_data[0] = ko['TMP_ijk'](lon=-180.)
        new_data[-1] = new_data[0]  # wrap in longitude with new values

        # check for and perform latitude interpolation
        if lat_check:
            new_data[:, 0] = ko['TMP_ijk'](lat=-90.)
            new_data[:, -1] = ko['TMP_ijk'](lat=90.)
        del ko['TMP_ijk'], ko['TMP']

        # store new variable data in file
        new_var[:] = new_data
        new_var.datascale, new_var.units =\
            gitm_varnames[var_dict[variable_name][0]][1:]

    # close file
    data_out.close()
    return cdf_filename


@np.vectorize
def dts_to_hrs(datetime_string, filedate):
    '''Get hours since midnight from datetime string'''

    return (datetime.strptime(datetime_string, '%Y-%m-%d %H:%M:%S'
                              ).replace(tzinfo=timezone.utc) -
            filedate).total_seconds()/3600.


def GITMbin_toCDF(file_dir, flag_2D=False):
    '''Collect data from all files found with file_prefix into netCDF4 files
    '''
    # If flag_2D is False, then 2D files are not present and
    # 2D variables will be calculated
    # Takes much longer if False.

    from os.path import basename, isfile

    ftic = perf_counter()
    files = sorted(glob(file_dir+'*.bin'))  # collect files
    if len(files) == 0:
        print('No unconverted files found.')
        return
    print('Checking for files that need converting...', end="")

    # initialize date values from first file
    file_name = basename(files[0])
    file_dt_str = file_name.split('_t')[-1].split('_')[0]  # keep only YYMMDD
    if int(file_dt_str[:2]) > 60:  # later than 1960
        string_date = '19' + file_dt_str[:2] + '-' + file_dt_str[2:4] + '-' +\
            file_dt_str[4:6]
    else:
        string_date = '20' + file_dt_str[:2] + '-' + file_dt_str[2:4] + '-' +\
            file_dt_str[4:6]
    filedate = datetime.strptime(string_date+' 00:00:00', '%Y-%m-%d %H:%M:%S'
                                 ).replace(tzinfo=timezone.utc)  # dt object

    # loop through all .bin files
    file_count = 0
    for file in files:
        file_tic = perf_counter()
        cdf_file = file.replace('.bin', '.nc')
        if isfile(cdf_file):
            continue  # do not convert if conversion already completed

        # read data and attributes from binary file
        file_count += 1
        filename, coords, variables, var_dict, attrs = GitmBin(file, flag_2D)
        time = np.array([attrs['time'][:19]])  # accurate to the sec
        coords['time'] = dts_to_hrs(time, filedate)
        attrs['filedate'] = filedate.strftime('%Y-%m-%d')  # store in attrs

        # perform data wrangling and write to cdf file
        cdf_file = gitm_toCDF(file, coords, variables, var_dict, attrs)
        print(f'Converted {file} to {cdf_file} in ' +
              f'{perf_counter()-file_tic:.5f}s')
    if file_count > 0:
        print(f'Files of pattern {file_dir+"*.bin"} converted to ' +
              f'cdf in {perf_counter()-ftic:.5f}s')
    else:
        print('All files are already converted.')
    return True



'''
Written by Rebecca Ringuette, 2021
'''
from datetime import datetime, timezone

# variable name in file: [standardized variable name, descriptive term, units]
model_varnames = {'b1x': ['B_1x', 'x component of the deviation from the ' +
                          'Earth dipole field', 0, 'GSM', 'car',
                          ['time', 'X', 'Y', 'Z'], 'nT'],
                  'b1y': ['B_1y', 'x component of the deviation from the ' +
                          'Earth dipole field', 0, 'GSM', 'car',
                          ['time', 'X', 'Y', 'Z'], 'nT'],
                  'b1z': ['B_1z', 'x component of the deviation from the ' +
                          'Earth dipole field', 0, 'GSM', 'car',
                          ['time', 'X', 'Y', 'Z'], 'nT'],
                  'bx': ['B_x', 'x component of the magnetic field', 0, 'GSM',
                         'car', ['time', 'X', 'Y', 'Z'], 'nT'],
                  'by': ['B_y', 'y component of the magnetic field', 0, 'GSM',
                         'car', ['time', 'X', 'Y', 'Z'], 'nT'],
                  'bz': ['B_z', 'z component of the magnetic field', 0, 'GSM',
                         'car', ['time', 'X', 'Y', 'Z'], 'nT'],
                  'e': ['u', 'Energy density', 0, 'GSM', 'car',
                        ['time', 'X', 'Y', 'Z'], 'J/m**3'],
                  'jx': ['J_x', 'x component of the current density', 0, 'GSM',
                         'car', ['time', 'X', 'Y', 'Z'], 'muA/m**2'],
                  'jy': ['J_y', 'y component of the current density', 0, 'GSM',
                         'car', ['time', 'X', 'Y', 'Z'], 'muA/m**2'],
                  'jz': ['J_z', 'z component of the current density', 0, 'GSM',
                         'car', ['time', 'X', 'Y', 'Z'], 'muA/m**2'],
                  'p': ['P', 'Pressure', 0, 'GSM', 'car',
                        ['time', 'X', 'Y', 'Z'], 'nPa'],
                  'rho': ['N_p', 'proton number density', 0, 'GSM', 'car',
                          ['time', 'X', 'Y', 'Z'], '10**6/cm**3'],  # Mp/cc
                  'ux': ['v_x', 'x component of velocity', 0, 'GSM', 'car',
                         ['time', 'X', 'Y', 'Z'], 'km/s'],
                  'uy': ['v_y', 'y component of velocity', 0, 'GSM', 'car',
                         ['time', 'X', 'Y', 'Z'], 'km/s'],
                  'uz': ['v_z', 'z component of velocity', 0, 'GSM', 'car',
                         ['time', 'X', 'Y', 'Z'], 'km/s'],
                  'th': ['theta_Btilt', 'Dipole tilt angle: positive for the' +
                         ' northern hemisphere dipole axis tilting towards ' +
                         'the Sun in northern hemisphere summer', 0, 'GSM',
                         'car', ['time'], 'radians']}


# coordinates stored as x, y, z in earth radii.
def MODEL():

    from kamodo import Kamodo
    import spacepy.pybats as sp
    from glob import glob
    import time
    from os.path import basename, isfile
    from numpy import array, NaN, unique, zeros, ravel, ndarray
    from numpy import min, max, floor, log
    from time import perf_counter
    import kamodo_ccmc.readers.reader_utilities as RU

    from readers.OCTREE_BLOCK_GRID._interpolate_amrdata import ffi
    from readers.OCTREE_BLOCK_GRID._interpolate_amrdata import lib

    class MODEL(Kamodo):
        '''SWMF model data reader for magnetosphere outputs.

        Inputs:
            file_dir: a string representing the file directory of the
                model output data.
                Note: This reader 'walks' the entire dataset in the directory.
            variables_requested = a list of variable name strings chosen from
                the model_varnames dictionary in this script, specifically the
                first item in the list associated with a given key.
                - If empty, the reader functionalizes all possible variables
                    (default)
                - If 'all', the reader returns the model_varnames dictionary
                    above for only the variables present in the given files.
            filetime = boolean (default = False)
                - If False, the script fully executes.
                - If True, the script only executes far enough to determine the
                    time values associated with the chosen data.
            printfiles = boolean (default = False)
                - If False, the filenames associated with the data retrieved
                    ARE NOT printed.
                - If True, the filenames associated with the data retrieved ARE
                    printed.
            gridded_int = boolean (default = True)
                - If True, the variables chosen are functionalized in both the
                    standard method and a gridded method.
                - If False, the variables chosen are functionalized in only the
                    standard method.
            verbose = boolean (False)
                - If False, script execution and the underlying Kamodo
                    execution is quiet except for specified messages.
                - If True, be prepared for a plethora of messages.
        All inputs are described in further detail in
            KamodoOnboardingInstructions.pdf.

        Returns: a kamodo object (see Kamodo core documentation) containing all
            requested variables in functionalized form.
            The run constants c, g, and R are saved in the kamodo.constants
            dictionary. R is the radius of the minimum inner boundary in R_E,
            c is the reduced speed of light, and g is the ratio of specific
            heats. Consult model documentation for more information.

        Notes and instructions:
        - The SWMF global magnetosphere outputs are given in one or more files
          per timestep in *.out files. The structure of these files are
          advantageous for efficient interpolation, so no file conversion is
          attempted.
        - The files are small and contain one time step per file, so
          interpolation method 1 is chosen. A custom interpolator is required.
        - The reader requires code in readers/OCTREE_BLOCK_GRID/:
          Compile the library using the command
            python interpolate_amrdata_extension_build.py
          inside the anaconda/miniconda3/miniforge3 environment for Kamodo.
          The shared library file name contains the python version and the name
          of the operation system, e.g.,
            _interpolate_amrdata.cpython-39-darwin.so
        - There is a pull request pending to implement sort_unstructured_data
          in the SpacePy software package. Until that PR is merged into
          SpacePy, you will need to manually modify the package. The
          lib/site-packages/spacepy/pybats/__init__.py script needs to be
          modified in your installation to not sort unstructured data in
          _read_Idl_bin() by default. See detailed instructions on the 2022 AGU
          poster https://doi.org/10.22541/essoar.167214301.16153548/v1 , at the
          bottom of the second to last column.
        '''
        def __init__(self, file_dir, variables_requested=[],
                     filetime=False, verbose=False, gridded_int=True,
                     printfiles=False, **kwargs):
            super(MODEL, self).__init__(**kwargs)
            self.modelname = 'SWMF_GM'
            t0 = perf_counter()

            # first, check for file list, create if DNE
            list_file = file_dir + self.modelname + '_list.txt'
            time_file = file_dir + self.modelname + '_times.txt'
            self.times, self.pattern_files = {}, {}
            if not isfile(list_file) or not isfile(time_file):
                files = sorted(glob(file_dir+'*.out')) +\
                    sorted(glob(file_dir+'*.outs'))  # .out or .outs
                self.filename = ''.join([f+',' for f in files])[:-1]
                patterns = unique([basename(f)[:10] for f in files])
                # get time grid from files
                dt = sp.IdlFile(files[0]).attrs['time']
                if dt is not None:  # filedate given not always at midnight
                    self.filedate = datetime.strptime(
                        dt.isoformat()[:10], '%Y-%m-%d').replace(
                        tzinfo=timezone.utc)
                else:
                    self.filedate = datetime(1900, 1, 1, tzinfo=timezone.utc)

                # establish time attributes from filenames
                for p in patterns:
                    # get list of files to loop through later
                    pattern_files = sorted(glob(file_dir+p+'*.out')) +\
                        sorted(glob(file_dir+p+'*.outs'))
                    self.pattern_files[p] = pattern_files
                    self.times[p] = {'start': [], 'end': [], 'all': []}

                    # loop through to get times, one file per time step
                    self.times[p]['start'] = array([RU.tstr_to_hrs(
                        sp.IdlFile(f).attrs['strtime'][:-1].replace(
                            'h', ':').replace('m', ':'))
                        for f in pattern_files])
                    self.times[p]['end'] = self.times[p]['start'].copy()
                    self.times[p]['all'] = self.times[p]['start'].copy()

                # create time list file if DNE
                RU.create_timelist(list_file, time_file, self.modelname,
                                   self.times, self.pattern_files,
                                   self.filedate)
            else:  # read in data and time grids from file list
                self.times, self.pattern_files, self.filedate, self.filename =\
                    RU.read_timelist(time_file, list_file)
            if filetime:
                return  # return times as is to prevent infinite recursion
            # only one pattern, so simplifying code
            p = list(self.pattern_files.keys())[0]

            # perform initial check on variables_requested list
            if len(variables_requested) > 0 and variables_requested != 'all':
                test_list = [value[0] for key, value in model_varnames.items()]
                err_list = [item for item in variables_requested if item not in
                            test_list]
                if len(err_list) > 0:
                    print('Variable name(s) not recognized:', err_list)
                for item in err_list:
                    variables_requested.remove(item)
                if len(variables_requested) == 0:
                    return

            # collect variable list
            mhd = sp.IdlFile(self.pattern_files[p][0])
            # mhd = sp.IdlFile(self.pattern_files[p][0],
            #                  sort_unstructured_data=False)
            file_variables = [key for key in mhd.keys() if key not in
                              ['grid', 'status', 'x', 'y', 'z']] + ['th']
            if len(variables_requested) > 0 and variables_requested != 'all':
                gvar_list = [key for key, value in model_varnames.items()
                             if value[0] in variables_requested and
                             key in file_variables]

                # check for variables requested but not available
                if len(gvar_list) != len(variables_requested):
                    err_list = [value[0] for key, value in
                                model_varnames.items() if value[0] in
                                variables_requested and key not in gvar_list]
                    if len(err_list) > 0:
                        print('Some requested variables are not available:',
                              err_list)
            else:  # only input variables on the avoid_list if requested
                gvar_list = [key for key in file_variables
                             if key in model_varnames.keys()]
                # returns list of variables included in data files
                if variables_requested == 'all':
                    self.var_dict = {value[0]: value[1:] for key, value in
                                     model_varnames.items() if key in
                                     gvar_list}
                    return

            # retrieve coordinate grids and run constants
            self._X = array(unique(mhd['x']))
            self._Y = array(unique(mhd['y']))
            self._Z = array(unique(mhd['z']))
            self.constants = {'c': mhd.meta['c'], 'g': mhd.meta['g']}
            # older runs may not have these
            if 'R' in mhd.meta.keys():
                self.constants['R'] = mhd.meta['R']
            if 'NX' in mhd.meta.keys():
                self.constants['NX'] = mhd.meta['NX']
            if 'NY' in mhd.meta.keys():
                self.constants['NY'] = mhd.meta['NY']
            if 'NZ' in mhd.meta.keys():
                self.constants['NZ'] = mhd.meta['NZ']

            # store kay for each variable desired
            self.variables = {model_varnames[var][0]: {
                'units': model_varnames[var][-1],
                'data': p} for var in gvar_list}

            # store a few items
            self.missing_value = NaN
            if verbose:
                print(f'Took {perf_counter()-t0:.6f}s to read in data')
            if printfiles:
                print(self.filename)
            self.octree = {p: {} for p in self.pattern_files.keys()}

            # register interpolators for each requested variable
            t_reg = perf_counter()
            # store original list b/c gridded interpolators change key list
            varname_list = [key for key in self.variables.keys()]
            for varname in varname_list:
                self.register_variables(varname, gridded_int, verbose=verbose)
            if verbose:
                print(f'Took {perf_counter()-t_reg:.5f}s to register ' +
                      f'{len(varname_list)} variables.')
            if verbose:
                print(f'Took a total of {perf_counter()-t0:.5f}s to kamodofy' +
                      f' {len(varname_list)} variables.')

        def setup_octree(self, x, y, z, verbose=False):
            '''
            This function requires _interpolate_amrdata*.so in
            readers/OCTREE_BLOCK_GRID/
            '''
            if verbose:
                tic = time.time()

            Ncell = int(len(x))

            if verbose:
                print("setup_octree: Ncell=", Ncell)

            if 'NX' in self.constants.keys():
                NX = int(self.constants['NX'])
            else:
                NX = 4  # minimum block size in BATSRUS is 4x4xc4 cells
            if 'NZ' in self.constants.keys():
                NY = int(self.constants['NY'])
            else:
                NY = 4
            if 'NZ' in self.constants.keys():
                NZ = int(self.constants['NZ'])
            else:
                NZ = 4

            N_blks = int(len(x)/(NX*NY*NZ))

            if 'R' in self.constants.keys():
                R = float(self.constants['R'])
            else:
                R = 3.0

            # initialize cffi arrays with Python lists may be slightly faster
            octree = {}
            x = ffi.new("float[]", list(x))
            y = ffi.new("float[]", list(y))
            z = ffi.new("float[]", list(z))
            # 2 elements per block, smallest blocks have 2x2x2=8 positions
            # at cell corners (GUMICS), so maximum array size is N/4
            # BATSRUS blocks have at least 4x4x4 cell centers
            octree['xr_blk'] = ffi.new("float[]", 2*N_blks)
            octree['yr_blk'] = ffi.new("float[]", 2*N_blks)
            octree['zr_blk'] = ffi.new("float[]", 2*N_blks)
            octree['x_blk'] = ffi.new("float[]", NX*N_blks)
            octree['y_blk'] = ffi.new("float[]", NY*N_blks)
            octree['z_blk'] = ffi.new("float[]", NZ*N_blks)
            octree['box_range'] = ffi.new("float[]", 7)
            octree['NX'] = ffi.new("int[]", [NX])
            octree['NY'] = ffi.new("int[]", [NY])
            octree['NZ'] = ffi.new("int[]", [NZ])

            N_blks = lib.xyz_ranges(Ncell, x, y, z, octree['xr_blk'],
                                    octree['yr_blk'], octree['zr_blk'],
                                    octree['x_blk'], octree['y_blk'],
                                    octree['z_blk'], octree['box_range'],
                                    octree['NX'], octree['NY'],
                                    octree['NZ'], 1)

            if N_blks < 0:
                print('NX:', list(octree['NX']))
                print('NY:', list(octree['NY']))
                print('NZ:', list(octree['NZ']))
                print('Y:', list(y[0:16]))
                print('Z:', list(z[0:16]))
                print('X_blk:', list(octree['x_blk'][0:16]))
                print('Y_blk:', list(octree['y_blk'][0:16]))
                print('Z_blk:', list(octree['z_blk'][0:16]))
                raise IOError("Block shape and size was not determined.")

            octree['N_blks'] = ffi.new("int[]", [N_blks])

            N_octree = int(N_blks*8/7)
            octree['octree_blocklist'] = ffi.new("octree_block[]", N_octree)

            dx_blk = zeros(N_blks, dtype=float)
            dy_blk = zeros(N_blks, dtype=float)
            dz_blk = zeros(N_blks, dtype=float)

            for i in range(0, N_blks):
                dx_blk[i] = octree['xr_blk'][2*i+1]-octree['xr_blk'][2*i]
                dy_blk[i] = octree['yr_blk'][2*i+1]-octree['yr_blk'][2*i]
                dz_blk[i] = octree['zr_blk'][2*i+1]-octree['zr_blk'][2*i]

            dxmin_blk = min(dx_blk)
            dxmax_blk = max(dx_blk)
            dymin_blk = min(dy_blk)
            dymax_blk = max(dy_blk)
            dzmin_blk = min(dz_blk)
            dzmax_blk = max(dz_blk)
            XMIN = octree['box_range'][0]
            XMAX = octree['box_range'][1]
            YMIN = octree['box_range'][2]
            YMAX = octree['box_range'][3]
            ZMIN = octree['box_range'][4]
            ZMAX = octree['box_range'][5]
            p1 = int(floor((XMAX-XMIN)/dxmax_blk+0.5))
            p2 = int(floor((YMAX-YMIN)/dymax_blk+0.5))
            p3 = int(floor((ZMAX-ZMIN)/dzmax_blk+0.5))
            while ((int(p1/2)*2 == p1)
                   & (int(p2/2)*2 == p2)
                   & (int(p3/2)*2 == p3)
                   ):
                p1 = int(p1/2)
                p2 = int(p2/2)
                p3 = int(p3/2)

            if verbose:
                print("XMIN, XMAX, YMIN, YMAX, ZMIN, ZMAX: ",
                      XMIN, XMAX, YMIN, YMAX, ZMIN, ZMAX)
                print("p1, p2, p3: ", p1, p2, p3, "dxmin,dymin,dzmin: ",
                      dxmin_blk, dymin_blk, dzmin_blk)

            MAX_AMRLEVEL = int(log((XMAX-XMIN)/(p1*dxmin_blk))/log(2.)+0.5)+1
            octree['MAX_AMRLEVEL'] = MAX_AMRLEVEL

            if verbose:
                print("MAX_AMRLEVEL:", MAX_AMRLEVEL)

            octree['numparents_at_AMRlevel'] = ffi.new("int[]",
                                                       N_blks*(MAX_AMRLEVEL+1))
            octree['block_at_AMRlevel'] = ffi.new("int[]",
                                                  N_blks*(MAX_AMRLEVEL+1))
            success = int(-1)
            success = lib.setup_octree(N_blks,
                                       octree['xr_blk'], octree['yr_blk'],
                                       octree['zr_blk'], MAX_AMRLEVEL,
                                       octree['box_range'],
                                       octree['octree_blocklist'], N_octree,
                                       octree['numparents_at_AMRlevel'],
                                       octree['block_at_AMRlevel'])

            if verbose:
                toc = time.time()
                print('setup_octree: ', toc-tic, 'seconds elapsed')

            return (octree)

        def register_variables(self, varname, gridded_int, verbose=False):
            '''Functionalizes the indicated dataset.'''

            # determine coordinate variables and xvec by coord list
            key = self.variables[varname]['data']
            coord_list = [value[5] for key, value in model_varnames.items()
                          if value[0] == varname][0]
            coord_dict = {'time': {'units': 'hr',
                                   'data': self.times[key]['all']}}
            gvar = [key for key, value in model_varnames.items() if
                    value[0] == varname][0]
            coord_str = [value[3]+value[4] for key, value in
                         model_varnames.items() if value[0] == varname][0]

            if len(coord_list) == 1:  # read all the values in
                self.variables[varname]['data'] = ravel([
                    sp.IdlFile(f).meta[gvar] for f in self.pattern_files[key]])
                self = RU.Functionalize_Dataset(
                    self, coord_dict, varname, self.variables[varname],
                    gridded_int, coord_str, interp_flag=0)
                return

            # remainder of logic is for time + 3D variables
            coord_dict['X'] = {'units': 'R_E', 'data': self._X}
            coord_dict['Y'] = {'units': 'R_E', 'data': self._Y}
            coord_dict['Z'] = {'units': 'R_E', 'data': self._Z}

            # all of data is time + 3D spatial
            def func(i):  # i is the file/time number
                # initialize octree instance for file if not already done
                if i not in self.octree[key].keys():
                    # files is closed after read in spacepy.pybats.IdlFile
                    mhd = sp.IdlFile(self.pattern_files[key][i])
                    # pull request pending to implement sort_unstructured_data
                    # lib/site-packages/spacepy/pybats/__init__.py needs to be
                    # modified in your installation to not sort unstructured
                    # data in _read_Idl_bin() by default
                    #  def _read_idl_bin(pbdat, header='units', start_loc=0,
                    #                     keep_case=True, headeronly=False,
                    #                     sort_unstructured_data=False):
                    # ...
                    #
                    # # Unstructured data can be in any order, so let's sort it
                    # if gtyp == 'Unstructured' and sort_unstructured_data:
                    #     gridtotal = np.zeros(npts)
                    #
                    # Once pull request is implemented in spacepy use keyword:
                    # mhd = sp.IdlFile(self.pattern_files[key][i],
                    #                  sort_unstructured_data=False)
                    x = array(mhd['x'])
                    y = array(mhd['y'])
                    z = array(mhd['z'])
                    # initialize octree object and store in dictionary
                    self.octree[key][i] = self.setup_octree(x, y, z,
                                                            verbose=verbose)

                # update which octree arrays the library points to
                lib.setup_octree_pointers(self.octree[key][i]['MAX_AMRLEVEL'],
                                          self.octree[key][i]['octree_blocklist'],
                                          self.octree[key][i]['numparents_at_AMRlevel'],
                                          self.octree[key][i]['block_at_AMRlevel'])
                # get data from file
                data = sp.IdlFile(self.pattern_files[key][i])[gvar]  # dmarray
                var_data = ffi.new("float[]", list(data))
                
                # assign custom interpolator: Lutz Rastaetter 2021
                def interp(xvec):
                    tic = perf_counter()
                    if not isinstance(xvec, ndarray):
                        xvec = array(xvec)

                    X, Y, Z = xvec.T  # xvec can be used like this
                    if not isinstance(X, ndarray):
                        X = array([X])
                        Y = array([Y])
                        Z = array([Z])

                    xpos = ffi.new("float[]", list(X))
                    ypos = ffi.new("float[]", list(Y))
                    zpos = ffi.new("float[]", list(Z))
                    npos = len(X)
                    return_data = list(zeros(npos, dtype=float))
                    return_data_ffi = ffi.new("float[]", return_data)

                    IS_ERROR = lib.interpolate_amrdata_multipos(
                        xpos, ypos, zpos, npos, var_data, return_data_ffi)
                    if IS_ERROR != 0:
                        print("Warning: SWMF/BATSRUS interpolation failed.")

                    return_data[:] = list(return_data_ffi)
                    toc = perf_counter()
                    if verbose:
                        print(varname, ' interp: ', toc-tic, 'seconds elapsed')

                    return return_data

                return interp

            # functionalize the variable dataset
            tmp = self.variables[varname]
            tmp['data'] = zeros((2, 2, 2, 2))  # saves execution time
            self = RU.Functionalize_Dataset(
                self, coord_dict, varname, tmp, gridded_int, coord_str,
                interp_flag=1, func=func, func_default='custom')
            return
    return MODEL



#import numpy as np
from numpy import transpose,zeros,array,append,float32,NAN,vectorize
from time import perf_counter
from netCDF4 import Dataset
#from astropy.constants import R_earth

from datetime import datetime,timezone,timedelta
from numpy import zeros, flip,array,float32,NaN

from glob import glob
#import kamodo_ccmc.readers.OpenGGCM.read_b_grids as rbg
#import kamodo_ccmc.readers.OpenGGCM.readmagfile3d as rmhd
import kamodo_ccmc.readers.OpenGGCM.readOpenGGCM as ropgm

from os.path import sep,isfile,isdir,dirname,exists
from os import remove

# this function takes timestamps with fields separated by ":" and seconds being a float including milliseconds
#@vectorize
def dts_to_hrs(datetime_string, filedate):
    '''Get hours since midnight from datetime string'''
    
    return (datetime.strptime(datetime_string, '%Y:%m:%d:%H:%M:%S.%f').replace(tzinfo=timezone.utc)\
            -filedate).total_seconds()/3600.
           
openggcm_gm_variable_to_GSE_factors={"bx":-1,"by":-1,"bz":1,
                             "bx1":-1,"by1":-1,"bz1":1,
                             "vx":-1,"vy":-1,"vz":1,
                             "ex":-1,"ey":-1,"ez":1,
                             "xjx":-1,"xjy":-1,"xjz":1,
                             "rr":1,"pp":1,"resis":1}
openggcm_gm_variable_units={"B_x":'nT',"B_y":'nT',"B_z":'nT',
                    "B1_x":'nT',"B1_y":'nT',"B1_z":'nT',
                    "E_xedges":'mV/m',"E_yedges":'mV/m',"E_zedges":'mV/m',
                    "V_x":'km/s',"V_y":'km/s',"V_z":'km/s',
                    "J_x":'muA/m^2',"J_y":'muA/m^2',"J_z":'muA/m^2',
                    "rho":'kg/m^3',"P":'pPa',"eta":'m^2/s'}
       
openggcm_gm_varnames={
    ### 3D spatial variables - to be aggregated into 4D ###
# B field on cell centers - used in online visualization
    'bx':['B_x','x component of magnetic field',0,'GSE','car',['time','x','y','z'],'nT'],
    'by':['B_y','y component of magnetic field',1,'GSE','car',['time','x','y','z'],'nT'],
    'bz':['B_z','z component of magnetic field',2,'GSE','car',['time','x','y','z'],'nT'],
    # these are B1_x, B1_y, B1_z used for file line tracing in online visualization/Kameleon
    'bx1':['B_xfaces','x component of magnetic field (on grid cell faces)',0,'GSE','car',['time','x_bxfacces','y_bxfacces','z_bxfacces'],'nT'],
    'by1':['B_yfaces','y component of magnetic field (on grid cell faces)',1,'GSE','car',['time','x_byfacces','y_byfacces','z_byfacces'],'nT'],
    'bz1':['B_zfaces','z component of magnetic field (on grid cell faces)',2,'GSE','car',['time','x_bzfacces','y_bzfacces','z_bzfacces'],'nT'],
    'ex':['E_xedges','x component of electric field (on grid cell edges)',3,'GSE','car',['time','x_exedges','y_exedges','z_exedges'],'mV/m'],
    'ey':['E_yedges','y component of electric field (on grid cell edges)',4,'GSE','car',['time','x_eyedges','y_eyedges','z_eyedges'],'mV/m'],
    'ez':['E_zedges','z component of electric field (on grid cell edges)',5,'GSE','car',['time','x_ezedges','y_ezedges','z_ezedges'],'mV/m'],
    'vx':['V_x','x component of plasma velocity',6,'GSE','car',['time','x','y','z'],'km/s'],
    'vy':['V_y','y component of plasma velocity',7,'GSE','car',['time','x','y','z'],'km/s'],
    'vz':['V_z','z component of plasma velocity',8,'GSE','car',['time','x','y','z'],'km/s'],
    'xjx':['J_x','x component of current density',9,'GSE','car',['time','x','y','z'],'muA/m**2'],
    'xjy':['J_y','y component of current density',10,'GSE','car',['time','x','y','z'],'muA/m**2'],
    'xjz':['J_z','z component of current density',11,'GSE','car',['time','x','y','z'],'muA/m**2'],
    'rr':['N','plasma number denstity (hydrogen equivalent)',12,'GSE','car',['time','x','y','z'],'1/cm**3'],
    'resis':['eta','resistivity',13,'GSE','car',['time','x','y','z'],'m**2/s'],
    'pp':['P','plasma pressure',14,'GSE','car',['time','x','y','z'],'pPa'],
    # 2D spatial variables aggregated into 3D
}


def openggcm_combine_magnetosphere_files(full_file_prefix,cadence=None,requested_variables=None,verbose=False):
    # file prefix includes everything including '3df' for magnetosphere and intended year,month,day (and hour)
    # read matching dates, times from 3df_list to generate list of raw files.
    # coadence: Default: None -- use all available times, otherwise use cadence as input starting with first time in simualtion.
    # requested_variables default: None -- convert all available 3D fields, other vise use only the requeted variables to generate smaller NetCDF files.
    
    tic=perf_counter()
    file_prefix,file_datetime = full_file_prefix.split('.3df')
    file_path=dirname(file_prefix)+sep

    
    if verbose:
        print('file_prefix: ',file_prefix)
        print('file_datetime: ',file_datetime)

    file_prefix=file_prefix+'.3df'
    
    file_date_time=file_datetime.split('_')[1:]
    file_year,file_mon,file_day=file_date_time[0].split("-")
    file_date_fmt=file_year+'/'+file_mon+'/'+file_day
    if verbose:
        print(file_path)

    # look for runme input file in the current or the parent directory
    runme_file=file_path+'/runme'
    if not exists(runme_file):
        runme_file=file_path+'/../runme'
        if not exists(runme_file):
            print("openggcm_to_cdf: required 'runme' file not found!")
            return False

    runme_file_object=open(runme_file,'r')
    runme_inputs=runme_file_object.readlines()
    runme_file_object.close()
        
    list_file=file_prefix+'_list'
    if not isfile(list_file):
        print("Required list file not found: ",list_file)
        return False
    
    f=open(list_file,'r')
    list_data=f.readlines()
    f.close()
    if verbose:
        print('file_date_time',file_date_time)
        print('len(file_date_time): ',len(file_date_time))

    added_time_at_beginning=0
    added_time_at_end=0

    
    reduced_list_data=[list_data[1]]
    if cadence is not None:
        file_seconds0=int(((list_data[1].split())[0])[-6:])
        print(file_seconds0,cadence)
        for file_str in list_data[2:]:
            file_seconds=int(((file_str.split())[0])[-6:])
            if (file_seconds-file_seconds0) >= cadence:
                #file_seconds0=file_seconds
                file_seconds0=file_seconds0+cadence
                reduced_list_data.append(file_str)
                print('appending '+file_str)
    else:
        reduced_list_data=list_data[1:]
    
    if len(file_date_time) == 1:
        file_str_list=[s for s in reduced_list_data if file_date_fmt in s]
        if verbose:
            print(file_str_list)
        if cadence is not None:
            file_seconds0=int(file_str_list[0][-6:])
            print(file_seconds0)
            file_str_list2=[file_str_list[0]]
            for file_str in file_str_list:
                file_seconds=int(file_str[-6:])
                if file_seconds >= cadence:
                    file_seconds0=file_seconds
#                    file_seconds0=file_seconds0+cadence # 
                    file_str_list2.append(file_str)
            file_str_list=file_str_list2
                
        # add first file for next day
        this_datetime=datetime.strptime(file_date_time[0]+' 00:00:00', 
                                              '%Y-%m-%d %H:%M:%S').replace(tzinfo=timezone.utc)
        next_datetime=this_datetime.add+timedelta(days=1)
        next_date_fmt=next_datetime.strftime('%Y/%m/%d')
        next_file_str_list=[s for s in reduced_list_data if next_date_fmt in s]
        if next_file_str_list is not []:
            file_str_list=file_str_list.append(next_file_str_list[0])
            added_time_at_end=1

    if len(file_date_time) > 1:
        file_hour=file_date_time[1]
        file_str_list=[s for s in reduced_list_data if file_date_fmt in s and " "+file_hour+":" in s]

        this_datetime=datetime.strptime(file_date_time[0]+' '+file_hour+':00:00',
                                        '%Y-%m-%d %H:%M:%S').replace(tzinfo=timezone.utc)
#
# wrap in time: add first file from next hour or day
#
        next_datetime=this_datetime+timedelta(hours=1)
        next_date_fmt=next_datetime.strftime('%Y/%m/%d')
        next_hour=next_datetime.strftime('%H')
        next_file_str_list=[s for s in reduced_list_data if next_date_fmt in s and " "+next_hour+":" in s]
        if len(next_file_str_list) > 0:
            file_str_list.append(next_file_str_list[0])
            added_time_at_end=1
            

    file_names_raw=[file_str.split(' ')[0] for file_str in file_str_list]
    filedate=datetime.strptime(file_date_fmt,'%Y/%m/%d').replace(tzinfo=timezone.utc)
    
    file_names=[]
    for ifile in range(len(file_names_raw)):
        filename=file_names_raw[ifile]
        if exists(file_path+filename):
            file_names.append(filename)

    
    # find grid files
    filename=file_names[0]
    file_seconds=filename[-6:]

    # look for two possible grid files in the current or the parent directory
    grid_file_name=filename[:-11]+".grid"
    grid_file_b_name=filename[:-11]+".grid"
    grid_file_b=None
    grid_file=file_path+'/'+grid_file_name
    xgrid_file=file_path+sep+'gridx.txt'
    ygrid_file=file_path+sep+'gridy.txt'
    zgrid_file=file_path+sep+'gridz.txt'
    if exists(xgrid_file) and exists(ygrid_file) and exists(zgrid_file):
        grid_file=[xgrid_file,ygrid_file,zgrid_file]
    else:            
        if not exists(grid_file):
            grid_file=file_path+'/../'+grid_file_name
            if not exists(grid_file):
                grid_file_=None
            grid_file_b=file_path+'/'+grid_file_b_name
        if grid_file is None:
            if not exists(grid_file_b):
                grid_file_b=file_path+'/../'+grid_file_b_name
                if exists(grid_file_b):
                    grid_file=grid_file_b

    if verbose:
        print('grid file with grid for b and e: ',grid_file_b)
    openggcm_missing=1.e30
            
    if grid_file is not None:
        if len(grid_file) > 1:
            # separate text files containing (x,dx,x_bx), (y,dy,y_by) or (z,dz,z_bz)
            # (x,y,z) in the model are (-x_GSE,-y_GSE,z_GSE)
            # using flip() to make arrays with increasing values in x and y
            gridx_file_object=open(xgrid_file)
            gridx_data=gridx_file_object.readlines()
            gridx_file_object.close()

            gx=-flip(array([float((line.split())[0]) for line in gridx_data[1:]]))
            nx=len(gx)

            gx_bx=-flip(array([float((line.split())[2]) for line in gridx_data[1:]]))
            gx_by=gx
            gx_bz=gx

            gx_ex=gx
            gx_ey=gx_bx
            gx_ez=gx_bx

            gridy_file_object=open(ygrid_file)
            gridy_data=gridy_file_object.readlines()
            gridy_file_object.close()
            
            gy=-flip(array([float((line.split())[0]) for line in gridy_data[1:]]))
            ny=len(gy)

            gy_bx=gy
            gy_by=-flip(array([float((line.split())[2]) for line in gridy_data[1:]]))
            gy_bz=gy

            gy_ex=gy_by
            gy_ey=gy
            gy_ez=gy_by


            gridz_file_object=open(zgrid_file)
            gridz_data=gridz_file_object.readlines()
            gridz_file_object.close()

            gz=array([float((line.split())[0]) for line in gridz_data[1:]])
            nz=len(gz)

            gz_bx=gz
            gz_by=gz
            gz_bz=array([float((line.split())[2]) for line in gridz_data[1:]])

            gz_ex=gz_bz
            gz_ey=gz_bz
            gz_ez=gz
            
        else:
                
            gx=zeros(5000)
            gy=zeros(5000)
            gz=zeros(5000)
            nx,ny,nz,gx,gy,gz = ropgm.read_grid_for_vector(grid_file,' ',gx,gy,gz)
            if verbose:
                print('NX: %i NY: %i N: %i' % (nx,ny,nz) )
            
                gx=-flip(gx[0:nx])
                gy=-flip(gy[0:ny])
                gz=gz[0:nz]
                gx_bx=zeros(nx)
                gy_bx=zeros(ny)
                gz_bx=zeros(nz)
                gx_by=zeros(nx)
                gy_by=zeros(ny)
                gz_by=zeros(nz)
                gx_bz=zeros(nx)
                gy_bz=zeros(ny)
                gz_bz=zeros(nz)
                gx_ex=zeros(nx)
                gy_ex=zeros(ny)
                gz_ex=zeros(nz)
                gx_ey=zeros(nx)
                gy_ey=zeros(ny)
                gz_ey=zeros(nz)
                gx_ez=zeros(nx)
                gy_ez=zeros(ny)
                gz_ez=zeros(nz)
                
                nx1,ny1,nz1,gx_bx,gy_bx,gz_bx = ropgm.read_grid_for_vector(grid_file,'bx',gx_bx,gy_bx,gz_bx)
                if nx1 <= 0 or ny1 <=0 or nz1 <=0:
                    raise IOError("bx grid not found")
                gx_bx=-flip(gx_bx[0:nx1])
                gy_bx=-flip(gy_bx[0:ny1])
                gz_bx=gz_bx[0:nz1]

                nx1,ny1,nz1,gx_by,gy_by,gz_by = ropgm.read_grid_for_vector(10,grid_file,'by',gx_by,gy_by,gz_by)
                if nx1 <= 0 or ny1 <=0 or nz1 <=0:
                    raise IOError("by grid not found")
                gx_by=-flip(gx_by[0:nx1])
                gy_by=-flip(gy_by[0:ny1])
                gz_by=gz_by[0:nz1]

                nx1,ny1,nz1,gx_bz,gy_bz,gz_bz = ropgm.read_grid_for_vector(10,grid_file,'bz',gx_bz,gy_bz,gz_bz)
                if nx1 <= 0 or ny1 <=0 or nz1 <=0:
                    raise IOError("bz grid not found")

                gx_bz=-flip(gx_bz[0:nx1])
                gy_bz=-flip(gy_bz[0:ny1])
                gz_bz=gz_bz[0:nz1]

                nx1,ny1,nz1,gx_ex,gy_ex,gz_ex = ropgm.read_grid_for_vector(grid_file,'ex',gx_ex,gy_ex,gz_ex)
                if nx1 <= 0 or ny1 <=0 or nz1 <=0:
                    raise IOError("ex grid not found")
                gx_ex=-flip(gx_ex[0:nx1])
                gy_ex=-flip(gy_ex[0:ny1])
                gz_ex=gz_ex[0:nz1]

                nx1,ny1,nz1,gx_ey,gy_ey,gz_ey = ropgm.read_grid_for_vector(10,grid_file,'ey',gx_ey,gy_ey,gz_ey)
                if nx1 <= 0 or ny1 <=0 or nz1 <=0:
                    raise IOError("ey grid not found")
                gx_ey=-flip(gx_ey[0:nx1])
                gy_ey=-flip(gy_ey[0:ny1])
                gz_ey=gz_ey[0:nz1]

                nx1,ny1,nz1,gx_ez,gy_ez,gz_ez = ropgm.read_grid_for_vector(10,grid_file,'ez',gx_ez,gy_ez,gz_ez)
                if nx1 <= 0 or ny1 <=0 or nz1 <=0:
                    raise IOError("ez grid not found")
                gx_ez=-flip(gx_ez[0:nx1])
                gy_ez=-flip(gy_ez[0:ny1])
                gz_ez=gz_ez[0:nz1]
            else:
                raise IOError("grid file not found")


    #initialize single output file
    nc_file=full_file_prefix+".nc"
    if exists(nc_file):
        remove(nc_file)
    else:
        print("The file %s does not exist" % nc_file) 
    
    # NetCDF root group
    data_out = Dataset(nc_file, 'w', format='NETCDF4')
    data_out.model = 'OpenGGCM_GM'
    data_out.file=''.join([f+',' for f in file_names]).strip(',')  #csv list of files
#    x_dim=data_out.create_dimension("x",len(gx))
#    y_dim=data_out.create_dimension("y",len(gy))
#    z_dim=data_out.create_dimension("z",len(gz))
#    time_dim=data_out.create_dimension("time",ntime)
    
    # store dimensions
    dim_dict={}
#    dim_dict['time']=array(file_times,dtype=float32)  # {'data':file_times,'datatype':float32,'size':len(file_times)}

    dim_dict['x']=array(gx,dtype=float32) # {'data':gx,'datatype':float32,'size':len(gx)}
    dim_dict['y']=array(gy,dtype=float32) # {'data':gy,'datatype':float32,'size':len(gy)}
    dim_dict['z']=array(gz,dtype=float32) # {'data':gz,'datatype':float32,'size':len(gz)}
    dim_dict['x_bxfaces']=gx_bx # {'data':gx_bx,'datatype':float32,'size':len(gx_bx)}
    dim_dict['y_bxfaces']=gy_bx # {'data':gy_bx,'datatype':float32,'size':len(gy_bx)}
    dim_dict['z_bxfaces']=gz_bx # {'data':gz_bx,'datatype':float32,'size':len(gz_bx)}
    dim_dict['x_byfaces']=gx_by # {'data':gx_by,'datatype':float32,'size':len(gx_by)}
    dim_dict['y_byfaces']=gy_by # {'data':gy_by,'datatype':float32,'size':len(gy_by)}
    dim_dict['z_byfaces']=gz_by # {'data':gz_by,'datatype':float32,'size':len(gz_by)}
    dim_dict['x_bzfaces']=gx_bz # {'data':gx_bz,'datatype':float32,'size':len(gx_bz)}
    dim_dict['y_bzfaces']=gy_bz # {'data':gy_bz,'datatype':float32,'size':len(gy_bz)}
    dim_dict['z_bzfaces']=gz_bz # {'data':gz_bz,'datatype':float32,'size':len(gz_bz)}
    dim_dict['x_exedges']=gx_ex # {'data':gx_ex,'datatype':float32,'size':len(gx_ex)}
    dim_dict['y_exedges']=gy_ey # {'data':gy_ex,'datatype':float32,'size':len(gy_ex)}
    dim_dict['z_exedges']=gz_ez # {'data':gz_ex,'datatype':float32,'size':len(gz_ex)}
    dim_dict['x_eyedges']=gx_ex # {'data':gx_ey,'datatype':float32,'size':len(gx_ey)}
    dim_dict['y_eyedges']=gy_ey # {'data':gy_ey,'datatype':float32,'size':len(gy_ey)}
    dim_dict['z_eyedges']=gz_ez # {'data':gz_ey,'datatype':float32,'size':len(gz_ey)}
    dim_dict['x_ezedges']=gx_ex # {'data':gx_ez,'datatype':float32,'size':len(gx_ez)}
    dim_dict['y_ezedges']=gy_ey # {'data':gy_ez,'datatype':float32,'size':len(gy_ez)}
    dim_dict['z_ezedges']=gz_ez # {'data':gz_ez,'datatype':float32,'size':len(gz_ez)}

    
#    return(nc_file,dim_dict)

    # the dimension variables do not include time (in hours) yet
    for dim in dim_dict.keys():
        new_dim = data_out.createDimension(dim, dim_dict[dim].size)
        new_var = data_out.createVariable('_'+dim, float32, dim) # dim_dict[dim]['datatype'],dim)
        new_var[:] = dim_dict[dim]
        if dim=='time': units = 'hr'
        else: units = 'R_E'
        new_var.units = units
        
    if verbose: print('Dimensions complete.\n')
#    data_out.close() 
#    return(nc_file,dim_dict)
   
    ntime=len(file_names)
    # variable loop
    #    print(nx,ny,nz,ntime)

    data=zeros(shape=(ntime,nx,ny,nz),dtype=float32,order='C')
#    fielddata=zeros(shape=(nx,ny,nz),dtype=float32,order='F')
    missing_value=NaN
    time=zeros(ntime)
    if requested_variables is None:
        convert_varnames=openggcm_gm_varnames
    else:
        convert_varnames=[]
        for varname in openggcm_gm_varnames:
            print(varname)
            if varname in requested_variables or openggcm_gm_varnames[varname][0] in requested_variables:
                convert_varnames.append(varname)                
        
#    for varname in ['bx']: #openggcm_gm_varnames:  #model_variables:
#    for varname in openggcm_gm_varnames:  #model_variables:
    for varname in convert_varnames:  #model_variables:
        variable_found=True
        data[:]=0.
        for ifile in range(ntime):
            filename=file_names[ifile]
            fielddata=zeros(shape=(nx,ny,nz),dtype=float32,order='F')
            fieldarray,nx_field,ny_field,nz_field,asciitime = ropgm.read_3d_field(file_path+filename,fielddata,varname);
            asciitime_str=asciitime.decode('utf-8')
            if verbose:
                print('asciitime_str: ',asciitime_str)
                print('time: ',time)
            try:
                timestr,time_seconds,time_epoch,time_str_colon,datatype = asciitime_str.split()
                if verbose:
                    print(time_str_colon)
                time[ifile]=dts_to_hrs(time_str_colon, filedate)
            except:
                if verbose:
                    print('empty asciitime string')
                variable_found=False
                break
            
            factor=openggcm_gm_variable_to_GSE_factors[varname]
#            fieldarray[(fieldarray == openggcm_missing)]=missing_value
            data[ifile,:,:,:]=factor*flip(flip(fieldarray,axis=0),axis=1)
        print(varname,data.min(),data.max(),variable_found)

        if variable_found:
        # add time in hours to dimensions
            dim='time'
            try:
                new_dim = data_out.createDimension(dim, len(time)) # should error if already there
                new_var = data_out.createVariable('_'+dim, float32, dim) # dim_dict[dim]['datatype'],dim)
                new_var[:] = time
                new_var.units = 'hr'
            except:
                if verbose:
                    print("time already in NetCDF file dimensions")

            dims = openggcm_gm_varnames[varname][-2]

            new_var = data_out.createVariable(varname, float32, dims)
            new_var[:] = data 
            new_var.units=openggcm_gm_varnames[varname][-1]

    runme_inputs=array(runme_inputs)
    runme_has_isphere=[line.find('ISPHERE') == 0 for line in runme_inputs]
    runme_isphere_line=(runme_inputs[runme_has_isphere])[0]
    isphere_name_value=runme_isphere_line.split()

    data_out.near_Earth_boundary_radius=float(isphere_name_value[1])
    data_out.near_Earth_boundary_radius_units='R_E'

    data_out.added_time_at_beginning=added_time_at_beginning
    data_out.added_time_at_end=added_time_at_end

    data_out.filedate=filedate.strftime("%Y-%m-%d")

    data_out.runme=runme_inputs
    print(f"Data for {full_file_prefix} converted in {perf_counter()-tic:.6f}s.")
    data_out.close() 

    toc=perf_counter()
    print('converted files in ',toc-tic,' seconds')                       
    return True




'''
Written by Rebecca Ringuette, 2021
'''

# variable name in file: [standardized variable name, descriptive term, units]
model_varnames = {'Ne': ['N_e', 'electron number density',
                         0, 'GDZ', 'sph', ['time', 'lon', 'lat', 'height'],
                         '1/m**3'],
                  'Te': ['T_e', 'electron temperature',
                         1, 'GDZ', 'sph', ['time', 'lon', 'lat', 'height'],
                         'K'],
                  'Ti': ['T_i', 'ion temperature',
                         2, 'GDZ', 'sph', ['time', 'lon', 'lat', 'height'],
                         'K'],
                  'Tn': ['T_n', 'neutral temperature',
                         3, 'GDZ', 'sph', ['time', 'lon', 'lat', 'height'],
                         'K'],
                  'O+': ['N_Oplus', 'number density of atomic oxygen ion',
                         4, 'GDZ', 'sph', ['time', 'lon', 'lat', 'height'],
                         '1/m**3'],
                  'H+': ['N_Hplus', 'number density of atomic hydrogen ion',
                         5, 'GDZ', 'sph', ['time', 'lon', 'lat', 'height'],
                         '1/m**3'],
                  'He+': ['N_Heplus', 'number density of atomic helium ion',
                          6, 'GDZ', 'sph', ['time', 'lon', 'lat', 'height'],
                          '1/m**3'],
                  'O2+': ['N_O2plus', 'number density of molecular oxygen ion',
                          7, 'GDZ', 'sph', ['time', 'lon', 'lat', 'height'],
                          '1/m**3'],
                  'NO+': ['N_NOplus', 'number density of molecular nitric ' +
                          'oxide', 8, 'GDZ', 'sph', ['time', 'lon', 'lat',
                                                     'height'], '1/m**3'],
                  'N+': ['N_Nplus', 'number density of atomic nitrogen ion',
                         9, 'GDZ', 'sph', ['time', 'lon', 'lat', 'height'],
                         '1/m**3'],
                  'TEC': ['TEC', 'vertical total electron content (height ' +
                          'integrated from bottom to top boundary)',
                          10, 'GDZ', 'sph', ['time', 'lon', 'lat'],
                          '10**16/m**2'],
                  'NmF2': ['NmF2', 'maximum electron number density in F2 ' +
                           'layer', 11, 'GDZ', 'sph', ['time', 'lon', 'lat'],
                           '1/m**3'],
                  'HmF2': ['HmF2', 'height of maximum electron number ' +
                           'density in F2 layer', 12, 'GDZ', 'sph',
                           ['time', 'lon', 'lat'], 'km']
                  }


# times from file converted to seconds since midnight of filedate
# plotting input times will be datetime strings of format 'YYYY-MM-DD HH:mm:ss'
# filedate is self.filedate from iri object
# converts to hours since midnight of filedate for plotting
def MODEL():

    from kamodo import Kamodo
    from netCDF4 import Dataset
    from glob import glob
    from os.path import basename, isfile
    from numpy import array, transpose, NaN, unique
    from numpy import where, append
    from time import perf_counter
    from datetime import datetime, timedelta, timezone
    import kamodo_ccmc.readers.reader_utilities as RU

    class MODEL(Kamodo):
        '''IRI model data reader.

        Inputs:
            file_dir: a string representing the file directory of the
                model output data.
                Note: This reader 'walks' the entire dataset in the directory.
            variables_requested = a list of variable name strings chosen from
                the model_varnames dictionary in this script, specifically the
                first item in the list associated with a given key.
                - If empty, the reader functionalizes all possible variables
                    (default)
                - If 'all', the reader returns the model_varnames dictionary
                    above for only the variables present in the given files.
            filetime = boolean (default = False)
                - If False, the script fully executes.
                - If True, the script only executes far enough to determine the
                    time values associated with the chosen data.
            printfiles = boolean (default = False)
                - If False, the filenames associated with the data retrieved
                    ARE NOT printed.
                - If True, the filenames associated with the data retrieved ARE
                    printed.
            gridded_int = boolean (default = True)
                - If True, the variables chosen are functionalized in both the
                    standard method and a gridded method.
                - If False, the variables chosen are functionalized in only the
                    standard method.
            verbose = boolean (False)
                - If False, script execution and the underlying Kamodo
                    execution is quiet except for specified messages.
                - If True, be prepared for a plethora of messages.
        All inputs are described in further detail in
            KamodoOnboardingInstructions.pdf.

        Returns: a kamodo object (see Kamodo core documentation) containing all
            requested variables in functionalized form.

        Notes:
            - IRI output files are given in multiple netCDF files per day.
            - The files are small and contain multiple time steps per file, so
              interpolation method 2 is chosen. The standard SciPy interpolator
              is used.
        '''
        def __init__(self, file_dir, variables_requested=[],
                     printfiles=False, filetime=False, gridded_int=True,
                     verbose=False, **kwargs):
            super(MODEL, self).__init__(**kwargs)
            self.modelname = 'IRI'
            t0 = perf_counter()

            # first, check for file list, create if DNE
            list_file = file_dir + self.modelname + '_list.txt'
            time_file = file_dir + self.modelname + '_times.txt'
            self.times, self.pattern_files = {}, {}
            if not isfile(list_file) or not isfile(time_file):
                # collect filenames
                files = sorted(glob(file_dir+'*.nc'))
                patterns = unique([basename(f)[:-10] for f in files])  # 2D, 3D
                self.filename = ''.join([f+',' for f in files])[:-1]

                # establish time attributes
                for p in patterns:
                    # get list of files to loop through later
                    pattern_files = sorted(glob(file_dir+p+'*.nc'))
                    self.pattern_files[p] = pattern_files
                    self.times[p] = {'start': [], 'end': [], 'all': []}

                    # loop through to get times
                    for f in range(len(pattern_files)):
                        cdf_data = Dataset(pattern_files[f])
                        tmp = array(cdf_data.variables['time'])/60. + \
                            float(f)*24.  # hrs since midnite 1st file
                        self.times[p]['start'].append(tmp[0])
                        self.times[p]['end'].append(tmp[-1])
                        self.times[p]['all'].extend(tmp)
                        cdf_data.close()
                    self.times[p]['start'] = array(self.times[p]['start'])
                    self.times[p]['end'] = array(self.times[p]['end'])
                    self.times[p]['all'] = array(self.times[p]['all'])

                # datetime object for midnight on date
                self.filedate = datetime(int(files[0][-10:-6]), 1, 1, 0, 0, 0
                                         ).replace(tzinfo=timezone.utc) + \
                    timedelta(days=int(files[0][-6:-3]) - 1)

                # create time list file if DNE
                RU.create_timelist(list_file, time_file, self.modelname,
                                   self.times, self.pattern_files,
                                   self.filedate)
            else:  # read in data and time grids from file list
                self.times, self.pattern_files, self.filedate, self.filename =\
                    RU.read_timelist(time_file, list_file)
            if filetime:
                return  # return times only

            # perform initial check on variables_requested list
            if len(variables_requested) > 0 and variables_requested != 'all':
                test_list = [value[0] for key, value in model_varnames.items()]
                err_list = [item for item in variables_requested if item not in
                            test_list]
                if len(err_list) > 0:
                    print('Variable name(s) not recognized:', err_list)
                for item in err_list:
                    variables_requested.remove(item)
                if len(variables_requested) == 0:
                    return

            # loop through file patterns for var mapping
            self.gvarfiles, self.varfiles, self.err_list = {}, {}, []
            for p in self.pattern_files.keys():
                # check var_list for variables not possible in this file set
                cdf_data = Dataset(self.pattern_files[p][0], 'r')
                if len(variables_requested) > 0 and\
                        variables_requested != 'all':
                    gvar_list = [key for key in model_varnames.keys()
                                 if key in cdf_data.variables.keys() and
                                 model_varnames[key][0] in variables_requested]
                    if len(gvar_list) != len(variables_requested):
                        err_list = [value[0] for key, value in
                                    model_varnames.items()
                                    if key not in cdf_data.variables.keys() and
                                    value[0] in variables_requested]
                        self.err_list.extend(err_list)  # add to master list
                else:
                    gvar_list = [key for key in model_varnames.keys()
                                 if key in cdf_data.variables.keys()]
                cdf_data.close()
                # store which file these variables came from
                self.varfiles[p] = [model_varnames[key][0] for
                                    key in gvar_list]
                self.gvarfiles[p] = gvar_list

            # clean up error list and then take action
            var_list = []
            for p in self.varfiles.keys():
                var_list.extend(self.varfiles[p])
            err_list = [var for var in self.err_list if var not in var_list]
            if len(err_list) > 0:
                print('Some requested variables are not available: ',
                      err_list)

            # collect all possible variables in set of files and return
            if variables_requested == 'all':
                self.var_dict = {value[0]: value[1:] for key, value in
                                 model_varnames.items() if value[0] in
                                 var_list}
                return

            # get coordinates and data from files
            self.variables = {}
            for p in self.pattern_files.keys():
                # get coordinates from first file of each type
                cdf_data = Dataset(self.pattern_files[p][0], 'r')
                lon = array(cdf_data.variables['lon'])
                lon_le180 = list(where(lon <= 180)[0])  # 0 to 180
                lon_ge180 = list(where((lon >= 180) & (lon < 360.))[0])
                tmp = lon - 180.  # now -180. to +180.
                setattr(self, '_lon_'+p, tmp)
                setattr(self, '_lon_idx_'+p, lon_ge180+lon_le180)
                setattr(self, '_lat_'+p,
                        array(cdf_data.variables['lat']))
                if '3D' in p:
                    setattr(self, '_height_'+p,
                            array(cdf_data.variables['ht']))
                cdf_data.close()
                # initialize variable dictionaries
                for var in self.gvarfiles[p]:
                    self.variables[model_varnames[var][0]] = {
                        'units': model_varnames[var][-1], 'data': p}

            # store a few items
            self.missing_value = NaN
            self._registered = 0
            if verbose:
                print(f'Took {perf_counter()-t0:.6f}s to read in data')
            if printfiles:
                print(self.filename)

            # register interpolators for each requested variable
            t_reg = perf_counter()
            # store original list b/c gridded interpolators change key list
            varname_list = list(self.variables.keys())
            for varname in varname_list:
                # determine which time grid applies
                self.register_variable(varname, gridded_int)
            if verbose:
                print(f'Took {perf_counter()-t_reg:.5f}s to register ' +
                      f'{len(varname_list)} variables.')
            if verbose:
                print(f'Took a total of {perf_counter()-t0:.5f}s to kamodofy' +
                      f' {len(varname_list)} variables.')

        # define and register a variable
        def register_variable(self, varname, gridded_int):
            """Registers an interpolator with proper signature"""

            # define and register the interpolators
            key = self.variables[varname]['data']
            coord_list = [value[5] for key, value in model_varnames.items()
                          if value[0] == varname][0]
            coord_dict = {'time': {'units': 'hr',
                                   'data': self.times[key]['all']}}
            if 'lat' in coord_list:   # 3D variables come from neutral file
                coord_dict['lon'] = {'units': 'deg', 'data':
                                     getattr(self, '_lon_'+key)}
                coord_dict['lat'] = {'units': 'deg', 'data':
                                     getattr(self, '_lat_'+key)}
            if 'height' in coord_list:
                coord_dict['height'] = {'units': 'km', 'data':
                                        getattr(self, '_height_'+key)}
            lon_idx = getattr(self, '_lon_idx_'+key)
            gvar = [key for key, value in model_varnames.items() if
                    value[0] == varname][0]  # variable name in file
            coord_str = [value[3]+value[4] for key, value in
                         model_varnames.items() if value[0] == varname][0]

            # define operations for each variable when given the key
            def func(i):
                '''key is the file pattern, start_idxs is a list of one or two
                indices matching the file start times in self.start_times[key].
                '''
                # get data from file
                file = self.pattern_files[key][i]
                cdf_data = Dataset(file)
                data = array(cdf_data.variables[gvar])
                if hasattr(cdf_data.variables[gvar][0], 'fill_value'):
                    fill_value = cdf_data.variables[gvar][0].fill_value
                else:
                    fill_value = None
                cdf_data.close()
                # if not the last file, tack on first time from next file
                if file != self.pattern_files[key][-1]:  # interp btwn files
                    next_file = self.pattern_files[key][i+1]
                    cdf_data = Dataset(next_file)
                    data_slice = array(cdf_data.variables[gvar][0])
                    cdf_data.close()
                    data = append(data, [data_slice], axis=0)
                # data wrangling
                if fill_value is not None:  # if defined, replace with NaN
                    data = where(data != fill_value, data, NaN)
                if len(data.shape) == 3:
                    # time, lat, lon -> time, lon, lat
                    variable = transpose(data, (0, 2, 1))
                elif len(data.shape) == 4:
                    # time, height, lat, lon -> time, lon, lat, height
                    variable = transpose(data, (0, 3, 2, 1))
                return variable[:, lon_idx]

            # functionalize the variable data (time chunked interpolation)
            self = RU.Functionalize_Dataset(
                self, coord_dict, varname, self.variables[varname],
                gridded_int, coord_str, interp_flag=2, func=func,
                times_dict=self.times[key])
            return
    return MODEL



'''
Spatial-only interpolation form:  Lutz Raestatter
Modified from spatial-only interpolation into this form:  Rebecca Ringuette
Kamodofication of the CTIPe model output

To check for pep8 standards:
    - pip install pycodestyle
    - pycodestyle ctipe_4D.py  # to show all errors
    - pycodestyle --first ctipe_4D.py  # to show first instances of error type
    - pycodestyle --show-source ctipe_4D.py  # to show full text per error
https://pypi.org/project/pycodestyle/ for more info.

filling on, default _FillValue of 9.969209968386869e+36 used
'''

# constants and dictionaries
model_varnames = {'density': ['rho_ilev1', 'total mass density', 0, 'GDZ',
                              'sph', ['time', 'lon', 'lat', 'ilev1'],
                              'kg/m**3'],
                  'density_2': ['rho', 'total mass density', 0, 'GDZ', 'sph',
                                ['time', 'lon', 'lat', 'height'], 'kg/m**3'],
                  'temperature': ['T_ilev1', 'temperature', 1, 'GDZ', 'sph',
                                  ['time', 'lon', 'lat', 'ilev1'], 'K'],
                  'temperature_2': ['T', 'temperature', 1, 'GDZ', 'sph',
                                    ['time', 'lon', 'lat', 'height'], 'K'],
                  'electron_temperature': ['T_e', 'electron temperature', 2,
                                           'GDZ', 'sph', ['time', 'lon', 'lat',
                                                          'height'], 'K'],
                  'ion_temperature': ['T_i', 'ion temperature', 3, 'GDZ',
                                      'sph', ['time', 'lon', 'lat', 'height'],
                                      'K'],
                  'height_n': ['H_ilev', 'height dependent on primary ' +
                               'pressure level', 4, 'GDZ', 'sph',
                               ['time', 'lon', 'lat', 'ilev'], 'm'],
                  'height_d': ['H_ilev1', 'height dependent on secondary ' +
                               'pressure level', 5, 'GDZ', 'sph',
                               ['time', 'lon', 'lat', 'ilev1'], 'm'],
                  'meridional_neutral_wind': ['v_nnorth_ilev', 'meridional ' +
                                              'neutral wind velocity (north)',
                                              6, 'GDZ', 'sph',
                                              ['time', 'lon', 'lat', 'ilev'],
                                              'm/s'],
                  'meridional_neutral_wind_2': ['v_nnorth', 'meridional ' +
                                                'neutral wind velocity ' +
                                                '(north)', 6, 'GDZ', 'sph',
                                                ['time', 'lon', 'lat',
                                                 'height'], 'm/s'],
                  'zonal_neutral_wind': ['v_neast_ilev', 'zonal neutral wind' +
                                         ' velocity (east)', 7, 'GDZ', 'sph',
                                         ['time', 'lon', 'lat', 'ilev'],
                                         'm/s'],
                  'zonal_neutral_wind_2': ['v_neast', 'zonal neutral wind ' +
                                           'velocity (east)', 7, 'GDZ', 'sph',
                                           ['time', 'lon', 'lat', 'height'],
                                           'm/s'],
                  'vertical_neutral_wind': ['v_nup_ilev', 'vertical neutral ' +
                                            'wind velocity (up)', 8, 'GDZ',
                                            'sph', ['time', 'lon', 'lat',
                                                    'ilev'], 'm/s'],
                  'vertical_neutral_wind_2': ['v_nup', 'vertical neutral ' +
                                              'wind velocity (up)', 8, 'GDZ',
                                              'sph', ['time', 'lon', 'lat',
                                                      'height'], 'm/s'],
                  'neutral_temperature': ['T_n_ilev', 'neutral temperature', 9,
                                          'GDZ', 'sph', ['time', 'lon', 'lat',
                                                         'ilev'], 'K'],
                  'neutral_temperature_2': ['T_n', 'neutral temperature', 9,
                                            'GDZ', 'sph',
                                            ['time', 'lon', 'lat', 'height'],
                                            'K'],
                  'mean_molecular_mass': ['m_avgmol_ilev1', 'mean molecular ' +
                                          'mass', 10, 'GDZ', 'sph',
                                          ['time', 'lon', 'lat', 'ilev1'],
                                          'amu'],
                  'mean_molecular_mass_2': ['m_avgmol', 'mean molecular mass',
                                            10, 'GDZ', 'sph',
                                            ['time', 'lon', 'lat', 'height'],
                                            'amu'],
                  'electron_density': ['N_e', 'electron number density', 11,
                                       'GDZ', 'sph',
                                       ['time', 'lon', 'lat', 'height'],
                                       '1/m**3'],
                  # 'N_n': ['N_n', 'variable description',
                  # 12, 'GDZ', 'sph', ['time', 'lon', 'lat', 'ilev'],'1/m**3'],
                  'solar_heating': ['Q_Solar_ilev', 'solar heating', 13, 'GDZ',
                                    'sph', ['time', 'lon', 'lat', 'ilev'],
                                    'J/kg/s'],
                  'solar_heating_2': ['Q_Solar', 'solar heating', 13, 'GDZ',
                                      'sph', ['time', 'lon', 'lat', 'height'],
                                      'J/kg/s'],
                  'joule_heating': ['Q_Joule_ilev', 'joule heating', 14, 'GDZ',
                                    'sph', ['time', 'lon', 'lat', 'ilev'],
                                    'J/kg/s'],
                  'joule_heating_2': ['Q_Joule', 'joule heating', 14, 'GDZ',
                                      'sph', ['time', 'lon', 'lat', 'height'],
                                      'J/kg/s'],
                  'radiation_heat_cool': ['Q_rad_ilev', 'radiative heating ' +
                                          'or cooling', 15, 'GDZ', 'sph',
                                          ['time', 'lon', 'lat', 'ilev'],
                                          'J/kg/s'],
                  'radiation_heat_cool_2': ['Q_rad', 'radiative heating or ' +
                                            'cooling', 15, 'GDZ', 'sph',
                                            ['time', 'lon', 'lat', 'height'],
                                            'J/kg/s'],
                  'atomic_oxygen_density': ['N_O_ilev', 'number density of ' +
                                            'atomic oxygen', 16, 'GDZ', 'sph',
                                            ['time', 'lon', 'lat', 'ilev'],
                                            '1/m**3'],
                  'atomic_oxygen_density_2': ['N_O', 'number density of ' +
                                              'atomic oxygen', 16, 'GDZ',
                                              'sph', ['time', 'lon', 'lat',
                                                      'height'], '1/m**3'],
                  'molecular_oxygen_density': ['N_O2_ilev', 'number density ' +
                                               'of molecular oxygen', 17,
                                               'GDZ', 'sph',
                                               ['time', 'lon', 'lat', 'ilev'],
                                               '1/m**3'],
                  'molecular_oxygen_density_2': ['N_O2', 'number density of ' +
                                                 'molecular oxygen', 17, 'GDZ',
                                                 'sph', ['time', 'lon', 'lat',
                                                         'height'], '1/m**3'],
                  'molecular_nitrogen_density': ['N_N2_ilev', 'number ' +
                                                 'density of molecular ' +
                                                 'nitrogen', 18, 'GDZ', 'sph',
                                                 ['time', 'lon', 'lat',
                                                  'ilev'], '1/m**3'],
                  'molecular_nitrogen_density_2': ['N_N2', 'number density ' +
                                                   'of molecular nitrogen',
                                                   18, 'GDZ', 'sph',
                                                   ['time', 'lon', 'lat',
                                                    'height'], '1/m**3'],
                  'nitric_oxide_density': ['N_NO_ilev', 'number density of ' +
                                           'molecular nitric oxide', 19, 'GDZ',
                                           'sph', ['time', 'lon', 'lat',
                                                   'ilev'], '1/m**3'],
                  'nitric_oxide_density_2': ['N_NO', 'number density of ' +
                                             'molecular nitric oxide', 19,
                                             'GDZ', 'sph', ['time', 'lon',
                                                            'lat', 'height'],
                                             '1/m**3'],
                  'nitric_oxide_ion_density': ['N_NOplus_ilev', 'number ' +
                                               'density of nitric oxide ion',
                                               20, 'GDZ', 'sph',
                                               ['time', 'lon', 'lat', 'ilev'],
                                               '1/m**3'],
                  'nitric_oxide_ion_density_2': ['N_NOplus', 'number density' +
                                                 ' of nitric oxide ion', 20,
                                                 'GDZ', 'sph',
                                                 ['time', 'lon', 'lat',
                                                  'height'], '1/m**3'],
                  'molecular_nitrogen_ion_density': ['N_N2plus_ilev',
                                                     'number density of ' +
                                                     'molecular nitrogen ion',
                                                     21, 'GDZ', 'sph',
                                                     ['time', 'lon', 'lat',
                                                      'ilev'], '1/m**3'],
                  'molecular_nitrogen_ion_density_2': ['N_N2plus',
                                                       'number density of ' +
                                                       'molecular nitrogen ' +
                                                       'ion', 21, 'GDZ', 'sph',
                                                       ['time', 'lon', 'lat',
                                                        'height'], '1/m**3'],
                  'molecular_oxygen_ion_density': ['N_O2plus_ilev', 'number ' +
                                                   'density of molecular ' +
                                                   'oxygen ion', 22, 'GDZ',
                                                   'sph', ['time', 'lon',
                                                           'lat', 'ilev'],
                                                   '1/m**3'],
                  'molecular_oxygen_ion_density_2': ['N_O2plus', 'number ' +
                                                     'density of molecular ' +
                                                     'oxygen ion', 22, 'GDZ',
                                                     'sph', ['time', 'lon',
                                                             'lat', 'height'],
                                                     '1/m**3'],
                  'atomic_nitrogen_ion_density': ['N_Nplus_ilev', 'number ' +
                                                  'density of atomic ' +
                                                  'nitrogen ion', 23, 'GDZ',
                                                  'sph', ['time', 'lon', 'lat',
                                                          'ilev'], '1/m**3'],
                  'atomic_nitrogen_ion_density_2': ['N_Nplus', 'number ' +
                                                    'density of atomic ' +
                                                    'nitrogen ion', 23, 'GDZ',
                                                    'sph', ['time', 'lon',
                                                            'lat', 'height'],
                                                    '1/m**3'],
                  'atomic_oxygen_ion_density': ['N_Oplus', 'number density ' +
                                                'of atomic oxygen ion', 24,
                                                'GDZ', 'sph',
                                                ['time', 'lon', 'lat',
                                                 'height'], '1/m**3'],
                  'atomic_hydrogen_ion_density': ['N_Hplus', 'number density' +
                                                  ' of atomic hydrogen ion',
                                                  25, 'GDZ', 'sph',
                                                  ['time', 'lon', 'lat',
                                                   'height'], '1/m**3'],
                  'pedersen_conductivity': ['sigma_P_ilev', 'Pedersen ' +
                                            'conductivity', 26, 'GDZ', 'sph',
                                            ['time', 'lon', 'lat', 'ilev'],
                                            'S/m'],
                  'pedersen_conductivity_2': ['sigma_P', 'Pedersen ' +
                                              'conductivity', 26, 'GDZ', 'sph',
                                              ['time', 'lon', 'lat', 'height'],
                                              'S/m'],
                  'hall_conductivity': ['sigma_H_ilev', 'Hall conductivity',
                                        27, 'GDZ', 'sph', ['time', 'lon',
                                                           'lat', 'ilev'],
                                        'S/m'],
                  'hall_conductivity_2': ['sigma_H', 'Hall conductivity', 27,
                                          'GDZ', 'sph', ['time', 'lon', 'lat',
                                                         'height'], 'S/m'],
                  'zonal_ion_velocity': ['v_inorth_ilev', 'meridional ion ' +
                                         'wind velocity (north)', 28, 'GDZ',
                                         'sph', ['time', 'lon', 'lat', 'ilev'],
                                         'm/s'],
                  'zonal_ion_velocity_2': ['v_inorth', 'meridional ion wind ' +
                                           'velocity (north)', 28, 'GDZ',
                                           'sph', ['time', 'lon', 'lat',
                                                   'height'], 'm/s'],
                  'meridional_ion_velocity': ['v_ieast_ilev', 'zonal ion ' +
                                              'wind velocity (east)', 29,
                                              'GDZ', 'sph',
                                              ['time', 'lon', 'lat', 'ilev'],
                                              'm/s'],
                  'meridional_ion_velocity_2': ['v_ieast', 'zonal ion wind ' +
                                                'velocity (east)', 29, 'GDZ',
                                                'sph', ['time', 'lon', 'lat',
                                                        'height'], 'm/s'],
                  'height_integrated_joule_heating': ['W_JouleH', 'height ' +
                                                      'integrated joule ' +
                                                      'heating', 30, 'GDZ',
                                                      'sph', ['time', 'lon',
                                                              'lat'],
                                                      'W/m**2'],
                  'energy_influx': ['Phi_E', 'energy flux', 31, 'GDZ', 'sph',
                                    ['time', 'lon', 'lat'], 'mW/m**2'],
                  'mean_energy': ['E_avg', 'average energy', 32, 'GDZ', 'sph',
                                  ['time', 'lon', 'lat'], 'keV'],
                  'total_electron_content': ['TEC', 'vertical total electron' +
                                             ' content (height integrated ' +
                                             'from bottom to top boundary)',
                                             33, 'GDZ', 'sph',
                                             ['time', 'lon', 'lat'], '1/m**2'],
                  # '10**16/m**2'
                  'theta_electric_field_at_140km': ['E_theta140km',
                                                    'Electric field at 140 ' +
                                                    'km, theta component', 34,
                                                    'GDZ', 'sph',
                                                    ['time', 'Elon', 'Elat'],
                                                    'V/m'],
                  'lambda_electric_field_at_140km': ['E_lambda140km',
                                                     'Electric field at 140 ' +
                                                     'km, lambda component',
                                                     35, 'GDZ', 'sph',
                                                     ['time', 'Elon', 'Elat'],
                                                     'V/m'],
                  'theta_electric_field_at_300km': ['E_theta300km',
                                                    'Electric field at 300 ' +
                                                    'km, theta component', 36,
                                                    'GDZ', 'sph',
                                                    ['time', 'Elon', 'Elat'],
                                                    'V/m'],
                  'lambda_electric_field_at_300km': ['E_lambda300km',
                                                     'Electric field at 300 ' +
                                                     'km, lambda component',
                                                     37, 'GDZ', 'sph',
                                                     ['time', 'Elon', 'Elat'],
                                                     'V/m']}


def MODEL():
    from numpy import array, NaN, unique, append, where, transpose
    from time import perf_counter
    from glob import glob
    from os.path import basename, isfile
    from datetime import datetime, timezone
    from kamodo import Kamodo
    from netCDF4 import Dataset
    import kamodo_ccmc.readers.reader_utilities as RU

    # main class
    class MODEL(Kamodo):
        '''CTIPe model data reader.

        Inputs:
            file_dir: a string representing the file directory of the
                model output data.
                Note: This reader 'walks' the entire dataset in the directory.
            variables_requested = a list of variable name strings chosen from
                the model_varnames dictionary in this script, specifically the
                first item in the list associated with a given key.
                - If empty, the reader functionalizes all possible variables
                    (default)
                - If 'all', the reader returns the model_varnames dictionary
                    above for only the variables present in the given files.
            filetime = boolean (default = False)
                - If False, the script fully executes.
                - If True, the script only executes far enough to determine the
                    time values associated with the chosen data.
            printfiles = boolean (default = False)
                - If False, the filenames associated with the data retrieved
                    ARE NOT printed.
                - If True, the filenames associated with the data retrieved ARE
                    printed.
            gridded_int = boolean (default = True)
                - If True, the variables chosen are functionalized in both the
                    standard method and a gridded method.
                - If False, the variables chosen are functionalized in only the
                    standard method.
            verbose = boolean (False)
                - If False, script execution and the underlying Kamodo
                    execution is quiet except for specified messages.
                - If True, be prepared for a plethora of messages.
        All inputs are described in further detail in
            KamodoOnboardingInstructions.pdf.

        Returns: a kamodo object (see Kamodo core documentation) containing all
            requested variables in functionalized form.

        Notes:
            - CTIPe output files are given in multiple netCDF files per day. No
              attempt is made to combine these files, but some pre-processing
              is needed to calculate the median height across the entire
              dataset (time, lon, lat) for a given pressure level value.
            - CTIPe data is given in several coordinate systems, two depending
              on pressure level - a unitless representation of height
              corresponding to a given atmospheric pressure. The preset values
              of pressure level in the two coordinate systems are not
              guaranteed to be identical, so they are inverted independently of
              the other unless only one is provided. In that case, the are
              assumed to be identical.
            - Pressure level inversion is performed in the reader_utilities
              script, specifically the PLevelInterp function. Two versions of
              all variables that depend on pressure level are created: the
              original and one dependent on height, which is created through
              Kamodo's function composition feature.
            - The files are small and contain multiple time steps per file, so
              interpolation method 2 is chosen. The standard SciPy interpolator
              is used.
        '''
        def __init__(self, file_dir, variables_requested=[],
                     filetime=False, printfiles=False, gridded_int=True,
                     verbose=False, **kwargs):

            # only the density, height and neutral files are combined
            super(MODEL, self).__init__()
            self.modelname = 'CTIPe'
            t0 = perf_counter()

            # first, check for file list, create if DNE
            list_file = file_dir + self.modelname + '_list.txt'
            time_file = file_dir + self.modelname + '_times.txt'
            self.times, self.pattern_files = {}, {}
            if not isfile(list_file) or not isfile(time_file):
                # collect filenames
                all_files = sorted(glob(file_dir+'*-plot-*.nc'))
                files = [f for f in all_files if 'plasma' not in f]
                patterns = unique([basename(f)[16:-3] for f in files])
                # e.g. 'density', 'neutral', 'height'
                self.filename = ''.join([f+',' for f in files])[:-1]
                # datetime object for midnight on date
                date = basename(files[0])[:10]  # 'YYYY-MM-DD'
                self.filedate = datetime.strptime(date+' 00:00:00',
                                                  '%Y-%m-%d %H:%M:%S').replace(
                                                      tzinfo=timezone.utc)
                filedate_ts = self.filedate.timestamp()

                # establish time attributes, using 3D time as default
                for p in patterns:
                    # get list of files to loop through later
                    pattern_files = sorted(glob(file_dir+'*'+p+'*.nc'))
                    self.pattern_files[p] = pattern_files
                    self.times[p] = {'start': [], 'end': [], 'all': []}

                    # loop through to get times
                    for f in range(len(pattern_files)):
                        cdf_data = Dataset(pattern_files[f])
                        tmp = array(cdf_data.variables['time'])  # utc tstamps
                        self.times[p]['start'].append(tmp[0])
                        self.times[p]['end'].append(tmp[-1])
                        self.times[p]['all'].extend(tmp)
                        cdf_data.close()
                    # convert timestamps to be hrs since midnight of 1st file
                    self.times[p]['start'] = (array(self.times[p]['start']) -
                                              filedate_ts)/3600.
                    self.times[p]['end'] = (array(self.times[p]['end']) -
                                            filedate_ts)/3600.
                    self.times[p]['all'] = (array(self.times[p]['all']) -
                                            filedate_ts)/3600.

                # perform post-processing to speed up pressure level conversion
                from kamodo_ccmc.readers.ctipe_tocdf import convert_all
                convert_all(file_dir, self.pattern_files, self.times)

                # create time list file if DNE
                RU.create_timelist(list_file, time_file, self.modelname,
                                   self.times, self.pattern_files,
                                   self.filedate)
            else:  # read in data and time grids from file list
                self.times, self.pattern_files, self.filedate, self.filename =\
                    RU.read_timelist(time_file, list_file)
            # return time info
            if filetime:
                return

            # collect variable dependency lists
            ilev1_list = [value[0] for key, value in model_varnames.items()
                          if value[5][-1] == 'ilev1']
            ilev1_replace = [item.split('_ilev1')[0] for item in ilev1_list if
                             item != 'H_ilev1']
            ilev_list = [value[0] for key, value in model_varnames.items() if
                         value[5][-1] == 'ilev']
            ilev_replace = [item.split('_ilev')[0] for item in ilev_list if
                            item != 'H_ilev']
            self.total_ilev = [item for item in ilev_list + ilev1_list if item
                               not in ['H_ilev', 'H_ilev1']]  # total ilev list
            self.total_replace = ilev_replace + ilev1_replace
            # dictionary mapping to navigate related variable names
            self.ilev_map = {item1: item2 for item1, item2 in
                             zip(self.total_replace, self.total_ilev)}

            # perform initial check on variables_requested list
            if (len(variables_requested) > 0) and (variables_requested !=
                                                   'all'):
                test_list = [value[0] for key, value in
                             model_varnames.items()]
                err_list = [item for item in variables_requested if item
                            not in test_list]
                if len(err_list) > 0:
                    print('Variable name(s) not recognized: ', err_list)
                for item in err_list:
                    variables_requested.remove(item)
                if len(variables_requested) == 0:
                    return

                # determine variable mapping
                # add ilev version of variables to the list, adding H_ilev(1)
                add_ilev = [var+'_ilev' for var in variables_requested if var
                            in ilev_replace]
                add_ilev1 = [var+'_ilev1' for var in variables_requested if var
                             in ilev1_replace]
                if len(add_ilev) > 0 or len(add_ilev1) > 0:  # add both
                    add_ilev += ['H_ilev']  # might need to replace one
                    add_ilev1 += ['H_ilev1']  # with the other
                new_var = variables_requested + add_ilev + add_ilev1
                short_var = [item for item in new_var if item not
                             in self.ilev_map.keys()]  # remove replaced items

            # collect variables per pattern type
            self.varfiles, self.gvarfiles = {}, {}
            self.err_list, self.var_dict = [], {}
            for p in self.pattern_files.keys():
                # check var_list for variables not possible in this file set
                cdf_data = Dataset(self.pattern_files[p][0], 'r')
                if len(variables_requested) > 0 and\
                        variables_requested != 'all':
                    gvar_list = [key for key, value in model_varnames.items()
                                 if value[0] in short_var and key in
                                 cdf_data.variables.keys()]  # file var names
                    # check for variables requested but not available in p file
                    if len(gvar_list) != len(short_var):
                        err_list = [value[0] for key, value in
                                    model_varnames.items() if value[0] in
                                    short_var and key not in
                                    cdf_data.variables.keys()]
                        # deal with height variable in two files (same name!)
                        if p == 'density':
                            if 'height' in cdf_data.variables.keys() and (
                                    'H_ilev1' in err_list):
                                gvar_list.append('height')
                                err_list.remove('H_ilev1')
                            if 'H_ilev' in err_list:  # remove neutral H
                                err_list.remove('H_ilev')
                        if p == 'neutral':
                            if 'height' in cdf_data.variables.keys() and (
                                    'H_ilev' in err_list):
                                gvar_list.append('height')
                                err_list.remove('H_ilev')
                            if 'H_ilev1' in err_list:  # remove density H
                                err_list.remove('H_ilev1')
                        if p == 'height':  # remove H from err list
                            if 'H_ilev' in err_list:
                                err_list.remove('H_ilev')
                            if 'H_ilev1' in err_list:
                                err_list.remove('H_ilev1')
                        if len(err_list) > 0:
                            self.err_list.extend(err_list)

                else:  # return full possible variable list
                    gvar_list = [key for key in cdf_data.variables.keys()
                                 if key in model_varnames.keys()]
                    # deal with height variable in two files (same name!)
                    if p in ['density', 'neutral'] and (
                            'height' in cdf_data.variables.keys()):
                        gvar_list.append('height')
                    if variables_requested == 'all':
                        var_dict = {value[0]: value[1:] for key, value in
                                    model_varnames.items() if key in gvar_list}
                        # deal with height again
                        if p == 'density' and 'height' in gvar_list:
                            var_dict['H_ilev1'] = \
                                model_varnames['height_d'][1:]
                        if p == 'neutral' and 'height' in gvar_list:
                            var_dict['H_ilev'] = \
                                model_varnames['height_n'][1:]
                        # add non-ilev versions of the variables in the files
                        key_list = list(var_dict.keys())
                        for var_key in key_list:
                            self.var_dict[var_key] = var_dict[var_key]
                            if var_key in self.total_ilev:
                                # retrieve equivalent non-ilev variable name
                                new_key = [key for key, value in
                                           self.ilev_map.items() if value ==
                                           var_key][0]
                                # retrieve key for model_varnames mapping
                                model_key = [key for key, value in
                                             model_varnames.items() if
                                             value[0] == new_key][0]
                                # add to returned dictionary
                                self.var_dict[model_varnames[model_key][0]] =\
                                    model_varnames[model_key][1:]
                        cdf_data.close()
                # store which file these variables came from
                self.varfiles[p] = [model_varnames[key][0] for key in gvar_list
                                    if key in model_varnames.keys()]
                if 'height' in gvar_list and p == 'density':
                    self.varfiles[p].append('H_ilev1')
                elif 'height' in gvar_list and p == 'neutral':
                    self.varfiles[p].append('H_ilev')
                self.gvarfiles[p] = gvar_list

            # return var_dict (already created)
            if variables_requested == 'all':
                return
            # clean up error list and then take action
            var_list = []
            for p in self.varfiles.keys():
                var_list.extend(self.varfiles[p])
            err_list = [var for var in self.err_list if var not in var_list]
            if len(err_list) > 0:
                print('Some requested variables are not available: ',
                      err_list)
            # perform height substitution if missing
            if 'H_ilev1' in self.err_list:  # density H missing
                self.gvarfiles['neutral'].append('height')
                print('Retrieving the H_ilev variable instead.')
            if 'H_ilev' in self.err_list:  # neutral H missing
                self.gvarfiles['density'].append('height')
                print('Retrieving the H_ilev1 variable instead.')

            # get coordinates from first file of each type
            self.variables = {}
            for p in self.pattern_files.keys():
                cdf_data = Dataset(self.pattern_files[p][0], 'r')
                lon = array(cdf_data.variables['lon'])
                lon_le180 = list(where(lon <= 180)[0])  # 0 to 180
                lon_ge180 = list(where(lon >= 180)[0])
                tmp = append(lon, 360.) - 180.  # now -180. to +180.
                setattr(self, '_lon_'+p, tmp)
                setattr(self, '_lon_idx_'+p, lon_ge180+lon_le180)
                setattr(self, '_lat_'+p,
                        array(cdf_data.variables['lat']))
                if p == 'density':
                    setattr(self, '_ilev1',
                            array(cdf_data.variables['plev']))
                    if isfile(file_dir+'CTIPe_km.nc'):  # km_ilev1 from file
                        km_data = Dataset(file_dir+'CTIPe_km.nc')
                        setattr(self, '_km_ilev1',
                                array(km_data.variables['km_ilev1']))
                        setattr(self, '_km_ilev1_max', km_data.km_ilev1_max)
                        setattr(self, '_km_ilev1_min', km_data.km_ilev1_min)
                        km_data.close()
                elif p == 'neutral':
                    setattr(self, '_ilev',
                            array(cdf_data.variables['plev']))
                    if isfile(file_dir+'CTIPe_km.nc'):  # km_ilev from file
                        km_data = Dataset(file_dir+'CTIPe_km.nc')
                        setattr(self, '_km_ilev',
                                array(km_data.variables['km_ilev']))
                        setattr(self, '_km_ilev_max', km_data.km_ilev_max)
                        setattr(self, '_km_ilev_min', km_data.km_ilev_min)
                        km_data.close()
                    lon = array(cdf_data.variables['elon'])  # 0 to 360
                    lon_le180 = list(where(lon <= 180)[0])  # 0 to 180
                    lon_ge180 = list(where((lon >= 180) & (lon < 360.))[0])
                    setattr(self, '_Elon_'+p, lon-180.)
                    setattr(self, '_Elon_idx_'+p, lon_ge180+lon_le180)
                    setattr(self, '_Elat_'+p,
                            array(cdf_data.variables['elat']))
                elif p == 'height':
                    setattr(self, '_height',
                            array(cdf_data.variables['ht']))
                cdf_data.close()

                # initialize variable dictionaries
                for var in self.gvarfiles[p]:
                    if var != 'height':
                        self.variables[model_varnames[var][0]] = {
                            'units': model_varnames[var][-1], 'data': p}
                    elif var == 'height' and p == 'density':
                        self.variables['H_ilev1'] = {
                            'units': model_varnames['height_d'][-1],
                            'data': p}
                    elif var == 'height' and p == 'neutral':
                        self.variables['H_ilev'] = {
                            'units': model_varnames['height_n'][-1],
                            'data': p}

            # print files to screen if option requested
            if printfiles:
                print('Files: \n',  self.filename)
            self.missing_value = NaN
            self._registered = 0

            # Check for presence of necessary height variables in varname_list.
            varname_list = [key for key in self.variables.keys()]
            ilev1_check = unique([True for item in varname_list if 'ilev1' ==
                                  item[-5:]])
            ilev_check = unique([True for item in varname_list if 'ilev' ==
                                 item[-4:]])
            if ilev1_check and 'H_ilev1' not in varname_list:
                self.ilev_sub = 'H_ilev1'
            elif ilev_check and 'H_ilev' not in varname_list:
                self.ilev_sub = 'H_ilev'
            else:
                self.ilev_sub = False

            # register interpolators for each requested variable
            # rearrange to deal with H_ilev and H_ilev1 first if there
            if 'H_ilev' in varname_list:
                varname_list.remove('H_ilev')
                varname_list = ['H_ilev'] + varname_list
            if 'H_ilev1' in varname_list:
                varname_list.remove('H_ilev1')
                varname_list = ['H_ilev1'] + varname_list
            t_reg = perf_counter()
            for varname in varname_list:
                self.register_variable(varname, gridded_int)
            if verbose:
                print(f'Took {perf_counter() - t_reg:.5f}s to register ' +
                      f'{len(varname_list)} variables.')
            if verbose:
                print(f'Took a total of {perf_counter() - t0:.5f}s to ' +
                      f'kamodofy {len(varname_list)} variables.')
            return

        # define and register a 3D variable
        def register_variable(self, varname, gridded_int):
            """Registers an interpolator with proper signature"""

            # determine coordinate variables and xvec by coord list
            key = self.variables[varname]['data']
            coord_list = [value[5] for key, value in model_varnames.items()
                          if value[0] == varname][0]
            coord_dict = {'time': {'units': 'hr',
                                   'data': self.times[key]['all']}}
            # special cases b/c in more than one file
            if (varname == 'N_e' and key != 'height') or \
                    (varname == 'm_avgmol_ilev1' and key != 'density'):
                key = 'neutral'
                coord_list = ['time', 'lon', 'lat', 'ilev']
            # retrieve coordinate grids
            if 'lat' in coord_list:
                coord_dict['lon'] = {'units': 'deg', 'data':
                                     getattr(self, '_lon_'+key)}
                coord_dict['lat'] = {'units': 'deg', 'data':
                                     getattr(self, '_lat_'+key)}
                lon_idx = getattr(self, '_lon_idx_'+key)
            if 'Elat' in coord_list:
                coord_dict['Elon'] = {'units': 'deg', 'data':
                                      getattr(self, '_Elon_'+key)}
                coord_dict['Elat'] = {'units': 'deg', 'data':
                                      getattr(self, '_Elat_'+key)}
                lon_idx = getattr(self, '_Elon_idx_'+key)
            if 'height' in coord_list:
                coord_dict['height'] = {'units': 'km', 'data':
                                        getattr(self, '_height')}
            if 'ilev' in coord_list:
                coord_dict['ilev'] = {'units': 'm/m', 'data':
                                      getattr(self, '_ilev')}
            if 'ilev1' in coord_list:
                coord_dict['ilev1'] = {'units': 'm/m', 'data':
                                       getattr(self, '_ilev1')}
            if varname not in ['H_ilev', 'H_ilev1']:
                gvar = [key for key, value in model_varnames.items() if
                        value[0] == varname][0]
            else:
                gvar = 'height'  # same in both files
            coord_str = [value[-4]+value[-3] for key, value in
                         model_varnames.items() if value[0] == varname][0]

            # define operations for each variable when given the key
            def func(i):
                '''key is the file pattern, start_idxs is a list of one or two
                indices matching the file start times in self.start_times[key].
                '''
                # get data from file
                file = self.pattern_files[key][i]
                cdf_data = Dataset(file)
                data = array(cdf_data.variables[gvar])
                if hasattr(cdf_data.variables[gvar][0], 'fill_value'):
                    fill_value = cdf_data.variables[gvar][0].fill_value
                else:
                    fill_value = None
                cdf_data.close()
                # if not the last file, tack on first time from next file
                if file != self.pattern_files[key][-1]:  # interp btwn files
                    next_file = self.pattern_files[key][i+1]
                    cdf_data = Dataset(next_file)
                    data_slice = array(cdf_data.variables[gvar][0])
                    cdf_data.close()
                    data = append(data, [data_slice], axis=0)
                # data wrangling
                if fill_value is not None:  # if defined, replace with NaN
                    data = where(data != fill_value, data, NaN)
                if len(data.shape) == 3 and 'Elon' not in coord_dict.keys():
                    variable = transpose(data, (0, 2, 1))
                elif 'Elon' in coord_dict.keys():
                    variable = data  # already correct dimension order
                elif len(data.shape) == 4:
                    variable = transpose(data, (0, 3, 2, 1))
                return variable[:, lon_idx]

            # define and register the interpolators, pull 3D data into arrays
            # need H functions to be gridded regardless of gridded_int value
            if varname in ['H_ilev', 'H_ilev1']:
                gridded_int = True
            self = RU.Functionalize_Dataset(
                self, coord_dict, varname, self.variables[varname],
                gridded_int, coord_str, interp_flag=2, func=func,
                times_dict=self.times[key])

            # perform substitution if needed
            if isinstance(self.ilev_sub, str) and varname == self.ilev_sub:
                other_name = ['H_ilev', 'H_ilev1']
                other_name.remove(varname)  # first element is the other name
                print(f'{other_name[0]} missing in data and is needed to ' +
                      'convert the requested variables to depend on height.' +
                      f' Using {varname} instead.')
                self.variables[other_name[0]] = self.variables[varname]
                # register the other variable by linking to this one
                self[varname] = other_name[0]

            # create pressure level -> km function once per ilev type
            if varname in ['H_ilev', 'H_ilev1'] or varname in self.total_ilev:
                if varname in ['H_ilev', 'H_ilev1']:  # create custom interp
                    new_varname = 'P'+coord_list[-1][1:]
                    kms = getattr(self, '_km_'+coord_list[-1])
                    kms_max = getattr(self, '_km_'+coord_list[-1]+'_max')
                    kms_min = getattr(self, '_km_'+coord_list[-1]+'_min')
                    # perform unit conversion if needed
                    if self.variables[varname]['units'] != 'km':
                        self[varname+'km_ijk[km]'] = varname + '_ijk'
                        km_interp = self[varname+'km_ijk']
                    else:
                        km_interp = self[varname+'_ijk']
                    # Import and call custom interpolator
                    units = 'm/m'
                    self[new_varname], interp_ijk = RU.PLevelInterp(
                        km_interp, coord_dict['time']['data'],
                        coord_dict['lon']['data'], coord_dict['lat']['data'],
                        coord_dict[coord_list[-1]]['data'], units, kms,
                        [kms_min, kms_max])
                    # kms is a 1D array of the median height values in km
                else:  # define by function composition
                    new_varname = varname.split('_ilev')[0]
                    units = self.variables[varname]['units']
                    # substitute kms array if height was also substituted
                    if isinstance(self.ilev_sub, str):
                        other_name = ['ilev', 'ilev1']
                        other_name.remove(self.ilev_sub[2:])
                        kms = getattr(self, '_km_'+other_name[0])
                        kms_max = getattr(self, '_km_'+other_name[0]+'_max')
                        kms_min = getattr(self, '_km_'+other_name[0]+'_min')
                        self[new_varname] = varname+'(P'+other_name[0][1:]+')'
                        interp_ijk = self[new_varname]
                        self[new_varname].meta['arg_units'] = \
                            self['P'+other_name[0][1:]].meta['arg_units']
                    else:
                        kms = getattr(self, '_km_'+coord_list[-1])
                        kms_max = getattr(self, '_km_'+coord_list[-1]+'_max')
                        kms_min = getattr(self, '_km_'+coord_list[-1]+'_min')
                        self[new_varname] = varname+'(P'+coord_list[-1][1:]+')'
                        interp_ijk = self[new_varname]
                        self[new_varname].meta['arg_units'] = \
                            self['P'+coord_list[-1][1:]].meta['arg_units']
                self.variables[new_varname] = {'units': units, 'data': key}

                # Register in kamodo object with a dedicated logic
                if gridded_int:
                    self = RU.register_griddedPlev(
                        self, new_varname, units, interp_ijk, coord_dict, kms,
                        [kms_min, kms_max])
            return

    return MODEL
